/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionDrawer;
import java.awt.Color;
import java.io.*;
import java.util.*;
import java.awt.event.*;

import BitmapFont.BitmapFontsFamily;

import gl4java.*;
import gl4java.awt.*;
import gl4java.utils.glut.*;
import gl4java.utils.glut.fonts.*;


/**
 * An abstract class representing a multi-scale rectangular grid on which 
 * shapes consisting of horizontal and vertical line segments (such as
 * trees and graphs) can be drawn
 * 
 * 
 * @author  Tamara Munzner, Serdar Tasiran
 * @version 
 * @see     AccordionDrawer.AccordionTreeDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 */

public abstract class AccordionDrawer extends GLCanvas implements MouseListener, MouseMotionListener, KeyListener {

//	protected SplitTree splitTree;
	static String fontFile = "/OlduvaiFont";
	static BitmapFontsFamily bff = new BitmapFontsFamily(fontFile);
	static boolean fontInitialized = false;
	public static int getDescent(int fontheight)
	{
		return bff.getDescent(fontheight);
	}
	public static int stringWidth(String name, int fontheight)
	{
		return bff.stringWidth(name, fontheight);
	}
	/**
	 * Initialize variable font family for tree labels. fontFamily[i]
	 * is the font should be used when there enough space for font
	 * with size i+minFontSize. The reason this is stored in an array
	 * is to speed up the look up time. 
	 * 
	 * @author Li Zhang
	 * @see TreeNode.drawInCell
	 */
	private void initializeFont() {
		// bitmap fonts
		try {
			bff.read();
		} catch (IOException e) {
			System.out.println("cannot produce fonts");
			return;
		}
		System.out.println("bitmap fonts initialized" + " " + bff.nofonts);
	}
	
    protected int count;
    private static int getTimeOfDayCounter;

    GLUTFunc glut;
    public GLCapabilities glc;

	// the 2 lines of splits in X and Y directions
	protected SplitLine[] splitLine = new SplitLine[2];

	protected Vector ToDrawQ;
	public boolean flag = false;
	public static boolean loaded = false; // don't draw until the data is loaded
	protected boolean startFrame = false;
	protected boolean continueFrame = false;
	
	// never called
	public void drawBackgroundBox(Color boxColor)
	{
		splitLine[X].drawBackground(boxColor, frameNum);
		splitLine[Y].drawBackground(boxColor, frameNum);
	}

    /**
    * Size of the window to draw the AccordionDrawer in
    * winsize[0] = w(idth)
    * winsize[0] = h(eight)
    */ 
    protected int winsize[] = new int[2];
    protected int nameStuckVal;

    protected long dynamicStart, continueStart;
	protected long dynamicTime;
	protected long now; 
	protected int frameNum;
    public int changedMarksFrame;
	public boolean rendermode = false;
	
	
    Color backgroundColor;
    Color labelColor;
    Color labelBackColor;
    Color labelHiColor;
    Color labelBackHiColor;
    protected Color rubberbandColor;

    // make sure that highlighted stuff is always visible, even it
    // might otherwise overlap with nonhighlighted stuff: always draw
    // it in front.

    float backplane;
    public float hiliteplane;
    float labelplane;
    float labelbgplane;
    protected float gridplane;
    protected float objplane;
    public float interactionplane;
	public boolean drawBackground;

    // cells
    public GridCell focusCell;
    public GridCell defaultFocusCell;
	/**
	 * The root GridCell that contains all others. It is at level (gridDepth-1).
	 * rootCell = grids[gridDepth-1][0][0]
	 */
	// it's the top level root of all quad trees in a quad grid cell system
	// it's the vertical binary tree root of all splitlines to choose a horizontal binary tree root in a binary cell system
	public GridCell rootCell;
	
	/**
	 * store lens pixel
	 */
	protected double xlenspixels[][];
	protected double ylenspixels[][];
	protected double zoompixels[];
	public Lens lens;
	public Lens oldLens;

	protected boolean doLensView;
	protected boolean mouseChanged;
	protected boolean justPressed;
	protected boolean doZoom;

	protected ArrayList drawnLabels = new ArrayList();
    public int[] gridDepth;
    /**
     * How many cells there are in the bottom level of the quadtree,
     * indexed by X and Y. 
     */
    public int bottomSize[] = new int[2];
    /**
     * minContext is the percentage of screen area guaranteed for smooshed things */
    public double minContextInside = .01; 
    public double minContextPeriphery = .01; 
    /**
     * How much to shrink a cell-to-cell area if it is smaller than 2*inflateIncr;
     */
    protected double shrinkFactor = .5; 

    public final static int X=0;
    public final static int Y=1;

    public Hashtable toMove;

    public final boolean doDoubleBuffer = true; // this is our intent to double buffer
    
    public boolean ignoreProgressive = false;
    
    protected boolean doublebuffer; // this gets set on/off with the gl state of the same name
    public boolean keepDrawing;
    public boolean keepMoving;
    
    //  set before drawing, the next draw will only output to file
    // file name is window title with incrementing counter, .eps)
    public boolean takeSnapshot;
    public final static String snapShotExtension = ".eps";
    public static String snapshotName = "";
    public BufferedWriter snapShotWriter = null;
    
    public void newSnapshotName(String base)
    {
    	String directory = ".";
    	try
		{
    		directory = new File(directory).getCanonicalPath();
		}
    	catch (IOException ioe)
		{
    		System.out.println("Error resolving canonical path for this current directory");
    		return;
		}
//    	directory = directory.substring(0, directory.lastIndexOf('/'));
//    	directory = directory.substring(0, directory.lastIndexOf('/'));
//    	.substring(0, directory.lastIndexOf('/')).substring(0, directory.lastIndexOf('/'));
    	

//    	if (snapShotWriter != null)
//    	{
//    		try
//			{
//    			snapShotWriter.close();
//			}
//    		catch (IOException ioe)
//			{
//    			System.out.println("Error: IOException trying to close an open writer: " + snapShotWriter.toString());
//			}
//    	}
    	base = base.replace(' ', '_').replace(':', '.');
    	snapshotName = directory + "/" + base + snapShotExtension;
    	boolean success = false;
    	int incrementNumber = 0;
    	BufferedReader br = null;
    	while (!success)
    	{
    		try
			{
    			br = new BufferedReader( new FileReader( snapshotName ) );
			}
    		catch (FileNotFoundException fnfe)
			{
    			System.out.println("ok, found a file: " + snapshotName);
    			success = true;
			}
    		if (!success)
    		{
    			try
				{
    				br.close();
				}
    			catch (IOException ioe)
				{
    				System.out.println("Error: IOException trying to close a known open file: " + snapshotName);
				}
    			snapshotName = directory + "/" + base + (incrementNumber++) + snapShotExtension;
    		}
    	}
    	try
		{
    		FileWriter fw = new FileWriter(snapshotName);
    		snapShotWriter = new BufferedWriter( fw );
		}
    	catch (IOException ioe)
		{
    		System.out.println("Error: IOException trying to open a file for writing: " + snapshotName);
		}
    	
    }
    

    protected float flashBoxWidth;
    protected boolean doBox;
    protected InteractionBox flashBox;
    protected InteractionBox flashBoxOld;
	public InteractionBox baseBox;
    protected CellGeom flashGeom;
    protected CellGeom flashGeomOld;
    protected Color flashCol;
    protected int flashX, flashY, flashXOld, flashYOld;

    public boolean showalg;
    public boolean colorgrid;
    public boolean biggrid;
    public boolean nogrid;
    public boolean nogeoms;
    public boolean noflash;
    public int expandleaves;
    public boolean showdiffs;
    private int linethickness;
    public boolean drawlabels;
    public boolean labelposright;
    public boolean labeldrawback;
    public boolean labelpopup;
    public boolean dimcolors;
    public boolean dimbrite;
    public boolean linkednav;
    public boolean labeltransp;
    public boolean cacherange;
    public boolean checktime;
    public boolean stoprendering;
    public boolean dumpstats;
    public boolean horiz; // if true, draw horizontal (like trees), false for sequences  :)
	public boolean drawsplits;
	
    /**
     * Order to draw elements in scene. 
     * BRUTEFORCE / 0: start from tree root, doesn't support highlighting
     * 2: FROMFOCUSCELL / 2: start from focusCell, add attached objects parents/children to queue
     */
    public int draworder;
    public static final int BRUTEFORCE = 0;
    public static final int FROMFOCUSCELL = 1;
    /**
       Whether to guarantee visibility. Default true.
    */
    public boolean guarvis;

    protected boolean mouseDragging;
    public boolean transitionAnimating;

    //boolean overlay;
    protected boolean doingFlash;

    int numAnimSteps; // in an animated transition
    protected double inflateIncr;
    
    // ST: Is this out of date?
    // array of array lists, one for each column of the bottom level 0
    // grid cells. the size of this array is equal to the depth of the
    // tree. each arraylist in this array has the bottom cells for
    // that column, which is a variable number since the grid is sparse. 
    // 
    // so we can access the hierarchical grid either from the top with
    // the rootCell or from the bottom with the level 0 cells. 

    double minCellDims[] = new double[2];

    int key;

    protected PrintWriter stats;

    int numCellsDrawnThisFrame;
	public int drawDirection;
//	public int upDown;
//	public static final int LEFT = 0, RIGHT = 1, UP = 2, DOWN = 3;
	public int[]  xInterval, yInterval;
    
	public static String[] stateMachine = {"GL_DEPTH_TEST"};
	public static boolean[] stateMachineOn = {false};
	
    public AccordionDrawer(int w, int h) {
    	
	super(w,h);
	System.out.println("new accordion drawer: width: " + w +" height: " + h);
	winsize[0] = w;
	winsize[1] = h;
	
	if (!fontInitialized) {
		initializeFont();
		fontInitialized = true;
	}
	
	//ToDrawQ = Collections.synchronizedSortedSet(new TreeSet());
	ToDrawQ = new Vector();
	
	frameNum = 0;
	changedMarksFrame = 0;
	minCellDims[X] = 1;
	minCellDims[Y] = 1;
//	setDoubleBuffer(doDoubleBuffer, true);

	GLContext.gljNativeDebug = false;
	GLContext.gljClassDebug = false;
	GLContext.gljThreadDebug = false;

	backgroundColor = new Color(1.0f, 1.0f, 1.0f, 0.0f);
	labelColor = new Color(0.0f, 1.0f, 0.0f, 0.0f);
	labelBackColor = new Color(0.2f, 0.2f, 0.2f, 0.0f);
	labelHiColor = new Color(0.2f, 0.2f, 0.2f, 0.0f);
	labelBackHiColor = new Color(0.0f, 1.0f, 0.0f, 0.0f);
	dynamicTime = 30;
	addMouseListener(this);
	addMouseMotionListener(this);
	addKeyListener(this);
	toMove = new Hashtable();
	keepDrawing = false;
	showalg = false;
	colorgrid = false;
	biggrid = true;
	nogrid = true;
	noflash = false;
	showdiffs = true;
	drawlabels = true;
	draworder = FROMFOCUSCELL;
	guarvis = true;
	labelposright = true;
	labeldrawback = false; // draw the background box of a label
	labelpopup = false;
	dimcolors = false;
	dimbrite = true;
	linkednav = true;
	labeltransp = false;
	checktime = true;
	stoprendering = false;

	doingFlash = false;
	mouseDragging = false;
	transitionAnimating = false;

	linethickness = 1;
	flashBoxWidth = 1;
	numAnimSteps = 10;
	inflateIncr = 0.15;
	gridplane = -1.0f;
	objplane = -.5f;
	hiliteplane = -.3f;
	labelbgplane = -.25f; 
	labelplane = -.2f; 
	interactionplane = objplane -.1f;

	count = 0;
	focusCell = null;
    }
    
    // the number of objects per pixel to guarantee gapless rendering
//    protected double pixelDiv = 1.0/10;
    
    /**
     * set the size of a block
     * @param newSize the new block size, in pixels, can be fractional
     */
    public void newPixelDiv(double newSize, int y)
    {
    	minCellDims[y] = newSize;
//    	pixelDiv = 1.0/newSize;
    }
    
	public double getPixelSize(int y)
	{
		double returnValue = s2w(minCellDims[y], y);
		if (returnValue <= 0)
			System.out.println("Error: pixel size is non-positive: " + returnValue);
		return returnValue;
	}
    
	public void endAllTransitions()
	{
		Enumeration enum = toMove.elements();
		while (enum.hasMoreElements())
		{
			SplitTransition st = (SplitTransition)enum.nextElement();
			st.end();
			toMove.remove(st);
		}
		transitionAnimating = false;
	}
    
	public void initSplitLines()
	{
		splitLine[X] = new SplitLine(this, true, SplitLine.defaultMinStuckValue, SplitLine.defaultMaxStuckValue,getBottomGrid(true)); // horizontal splitline
		splitLine[Y] = new SplitLine(this, false, SplitLine.defaultMinStuckValue, SplitLine.defaultMaxStuckValue, getBottomGrid(false)); // vertical splitline
	}
	
	/**
	  * clear all the data structures initiated in AD, called when tree(s) get deleted from AD
	  * @see TreeJuxtaposer.TreeJuxtaposer.deleteTree
	  *
	  */   
		public void shutdown() {				
				
				ToDrawQ=null; 		
				 focusCell = null;
				 defaultFocusCell=null;	
				 rootCell = null;
		
			   toMove=null;
			   flashBox=null; 
			   flashBoxOld=null;
			   flashGeom=null;
			   flashGeomOld=null;
			   stats=null;	
			   drawnLabels.clear();
			   drawnLabels = null;					 
		
			   bottomSize = null;		
			   minCellDims = null;		   
			  System.out.println("clean accordionDrawer");
		 }
      
		 protected void finalize() throws Throwable {		 
					 try {
						 shutdown();
					 }
					 finally {
//						System.out.println("finally clean the AccordionDrawer");
//						 super.finalize();     
					 }
		 }
    
    public void setKey(int i) { key = i;}
    public int getKey() { return key;}
    public Object getGraphicsStuff() {
	return gl;
    }

    
	///////////////////////////////////////////////////////////////////////////
	// init function
	
    public void init() 
    {
		float thecol[] = new float[3];
		backgroundColor.getRGBColorComponents(thecol);
		
		glc = glj.getGLCapabilities();
		gl.glClearColor( thecol[0], thecol[1], thecol[2], 1.0f);
		
		//gl.glEnable(GL_DEPTH_TEST);
		//gl.glDepthFunc(GL_LEQUAL);
		// leave depthfunc at LEQUAL all the time so that label
		// backgrounds and label strings can both be in the same
		// plane. no need to have it be LESS
		glut = new GLUTFuncLightImplWithFonts(gl, glu);
		gl.glLogicOp(GLEnum.GL_XOR);
		reshape(winsize[0],winsize[1]); // no redraw yet (keep manualReshape false)
		bff.setupGL(gl);
		System.out.println("in init!! " + winsize[0] + " " + winsize[1]);
    }
    
	  ///////////////////////////////////////////////////////////////////////////
	  // display function
	  public void display() {
	  	if (winsize[X] < 0 || winsize[Y] < 0)
	  	{
	  		System.out.println("don't draw when the size is negative!");
	  		return;
	  	}
//if (false) // cancel the direct call to display?
		redraw();
    	}

    
    /**
     * Reshapes the drawing area to new width w and height h in pixels
     *
     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
     * 
     *
     * @see      AccordionDrawer.GridCell
     */

	 public boolean manualReshape = false;
    public void reshape(int w, int h) {

    	if (w < 0 || h < 0)
    	{
    		System.out.println("either width or height is < 0: " + w + " " + h);
    		return;
    	}
    	if (w != winsize[X] || h != winsize[Y])
    	{
    		manualReshape = true;
//    		w = winsize[X];
////    		h = winsize[Y];
//    		System.out.println("debug");
//    		try{
//    		throw new Exception();
//    		}
//    		catch(Exception e)
//			{
//    			e.printStackTrace();
//			}
    	}
	System.out.println("reshape: " + w + " " + h);
	winsize[0] = w;
	winsize[1] = h;
	
	int gridMax = gridDepth[0] > gridDepth[1] ? gridDepth[0] : gridDepth[1];
	
	gl.glViewport(0, 0, winsize[0], winsize[1]);
	gl.glMatrixMode(GL_PROJECTION);
	gl.glLoadIdentity();
	gl.glOrtho( 0.0f, 1.0f, 0.0f, 1.0f, .01f, gridMax+4.0f );
	gl.glScalef(1.0f, -1.0f, 1.0f);
	gl.glTranslatef(0.0f, -1.0f, 0.0f);
	gl.glMatrixMode(GL_MODELVIEW);
	gl.glLoadIdentity();
//	if (manualReshape)
		requestRedraw();
//	else
//		System.out.println("reshape but no redraw");
	manualReshape = false;
    }

    /**
     * Redraws the entire canvas
     *
     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
     * 
     *
     * @see      AccordionDrawer.GridCell
     */
    public void redraw() { 
    	
    incrementFrameNumber(); // necessary for application forced repaints
	if (!loaded || !drawStart(false)) return; // not loaded or no swap
	if (keepDrawing){// && !keepMoving) {
	    continueFrame();
	} else{// if (!keepMoving) {
	    startNewFrame();
	}
	drawEnd(startFrame); // swap only after first frame //getDoubleBuffer());
    }


    /** Redraw immediately, do not continue filling in details.
     * @author Tamara Munzner */
    public void requestRedraw()
    {
//	System.out.println("redrawing movement: " + keepMoving + " framenum: " + frameNum);
    	incrementFrameNumber();
    	if (!showalg) keepDrawing = false;
	
	// glj is not null (sometimes) when this is called explicitly by us, but is null when:
	//  - gl window opens (is created, deiconified)
	//  - we add a drawer to the canvas (or maybe the canvas is displayed the first time)
	//  - mouse entries (perhaps related to the first time, not every time)
	//  - canvas gains focus (again, first time)
	// so, to make things robust, we need to check/free it either here, or before we get here,
	//  to make sure there are no external events that may cause us to redraw (perhaps a drawing pass that
	//  doesn't quite make it to the gljfree command will cause instability by leaving glj around)
    	if (glj != null)
    		glj.gljFree();
		repaint();
    }

    /** Enqueue a redraw request, but don't stop what we're in the
     * middle of doing. 
     * @author Tamara Munzner */
    public void requestRedrawEventually() {
//    	System.out.println("rre");
	repaint();
    }

    public void setDoubleBuffer(boolean value, boolean clear) {
    	doublebuffer = value;
//System.out.println("setting double buffer: " + value + " and clearing: " + clear);
    	if (doublebuffer)
		{
    		if (startFrame)
    		{
//    			System.out.println("GL_BACK");
    			gl.glDrawBuffer(GL_BACK);
    		}
    		else if (continueFrame)
    			// the first continue frame call to this function leaves startframe on
    		{
//    			System.out.println("GL_FRONT");
    			gl.glDrawBuffer(GL_FRONT);
    		}
//    		else
//    		{
////    			System.out.println("not start, not first continue");
//    		}
			if (glc != null)
			{
//				System.out.println("turning on double buffered");
				glc.setDoubleBuffered(true);
			}
			if (clear)
			{
				gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			}
		}
		else
		{ // no double buffer, for drawing during flash/rubberband
			gl.glDrawBuffer(GL_FRONT);
			if (glc != null)
				glc.setDoubleBuffered(false);
else System.out.println("glc is null, not setting new db value");
			if (clear)
				gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		}
    }
    public boolean getDoubleBuffer() { return doublebuffer;}

	public abstract void drawFrame();
	public abstract void drawRange(int min, int max);
	public abstract void drawDensity(int min, int max);

    /**
     * Starts a new frame and draws the contents of the
     * AccordionDrawer (enclosed in the rootCell) in a new frame.
     *
     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
     * 
     *
     * @see      AccordionDrawer.GridCell
     */

    void startNewFrame() {
		drawnLabels.clear();
		flashGeomOld = null; 
		flashBoxOld = null; 

		// set flag for first frame atrifact
		flag = false;
		
		// we need these flags for later on to
		// draw pixels later on in the correct buffer
		startFrame = true;
		continueFrame = false;

		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		// all rendering state changes should be set
		// locally - means: never put a rendering state
		// change anywhere in the code if you don't want
		// to draw something at this point
		// once you finished rendering, change the state back
		// to the original state
		// see InteractionBox.java for instance
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		dynamicStart = System.currentTimeMillis();
		continueStart = dynamicStart;
		drawPreNewFrame();
		seedQueue();
		setDoubleBuffer(doDoubleBuffer, !takeSnapshot); // this does a clear when not taking a picture
		drawFrame();
    }
    
    public void setFlashBox(InteractionBox interactionBox)
    {
    	flashBox = interactionBox;
    }

    protected void continueFrame() {
//    	while(keepDrawing && !keepMoving)
    	{
    		continueFrame = true;
    		if (startFrame) // first continue frame?
    		{
    			startFrame = false;
    			setDoubleBuffer(doDoubleBuffer, false);
    		}
			dynamicStart = System.currentTimeMillis();
			if (dumpstats) stats.print(" inbetween "+(dynamicStart-now));
			
			drawPreContFrame();
			drawFrame();

    	}
    }


    private void doShowAlg(GridCell c) {
		// delete for now - reimplmentation later on
    }


	/*
	 *  to clear the last frame
	 */
	public void clearLens()
	{
		if (null == glj || !glj.gljMakeCurrent())
			return;
		
//		int diffPos[] = new int[2];
//		diffPos[X] = lens.pos[X] - lens.oldPos[X];
//		diffPos[Y] = lens.pos[Y] - lens.oldPos[Y];
//		
//		gl.glDrawBuffer(GL_FRONT);
//		glc.setDoubleBuffered(false);
//		gl.glDepthMask(false);
//					
//		for(int i = 0; i <= lens.size[Y]; i++)
//		{
//			gl.glRasterPos3d(s2w(lens.oldPos[X] - lens.size[X]/2, X), s2w(lens.oldPos[Y] + lens.size[Y]/2 -i, Y), labelplane );
//			gl.glDrawPixels(lens.size[X], 1, GLEnum.GL_RGB, GLEnum.GL_FLOAT, xlenspixels[i]);
//		}
//		gl.glFlush();
//		glj.gljFree();

			
	}

	/**
		* render lens in a special mode
		* readPixels form the front buffer in x and y direction
		* to avoid flickering and overwrite only diffPos in x and y
		* 
		* to render the lens do a twolinesmove with picked cells
		* render the solution in a small viewport in the back buffer
		* read pixels of this viewport 
		* switch back in the original viewport 
		* draw stored pixels
		* 
		* TODO: x-distortion, check boundaries, flowmenu like zooming
		* 	
		* 
		*@author Kristian Hildebrand
		*/
		public void doLens() {
//			// 
//			double a = 0.45f;// minStuckVal[X];
//			double b = 0.55f;// maxStuckVal[X];
//			int pos[] = new int[2];
//			int diffPos[] = new int[2];
//			int drawPos[] = new int[2];
//			int drawSize[] = new int[2];
//		
//			if (null == glj || !glj.gljMakeCurrent())
//				return;
//
//			//System.out.println("lens at work... ");
//			if(!justPressed){
//			
//				gl.glDrawBuffer(GL_FRONT_AND_BACK);
//				gl.glDepthMask(false);
//					
//				// difference between mousepos between frames
//				diffPos[X] = lens.pos[X] - lens.oldPos[X];
//				diffPos[Y] = lens.pos[Y] - lens.oldPos[Y];
//				// divide four different cases to decide which pixels 
//				// on the x and y side have to be 'reconstructed' after
//				// the lens is moved
//				if(diffPos[Y] < 0) {					
//					//System.out.println("y less ");
//					for(int i = 0; i <= Math.abs(diffPos[Y]); i++) {
//						gl.glRasterPos3d(s2w(lens.oldPos[X] - lens.size[X]/2, X), s2w(lens.oldPos[Y] + lens.size[Y]/2 - i, Y), labelplane );
//						gl.glDrawPixels(lens.size[X], 1, gl.GL_RGB, gl.GL_FLOAT, xlenspixels[i]);
//					}
//				}
//				if(diffPos[Y] >= 0){	
//					//System.out.println("y more");				
//					for(int i = 0; i <= Math.abs(diffPos[Y]); i++){
//						gl.glRasterPos3d(s2w(lens.oldPos[X] - lens.size[X]/2, X), s2w(lens.oldPos[Y] - lens.size[Y]/2 + i, Y), labelplane );
//						gl.glDrawPixels(lens.size[X], 1, gl.GL_RGB, gl.GL_FLOAT, xlenspixels[lens.size[Y] - i]);
//					}
//				}
//				if(diffPos[X] < 0) {	
//					//System.out.println("x less");				
//					for(int i = 0; i <= Math.abs(diffPos[X]); i++){
//						gl.glRasterPos3d(s2w(lens.oldPos[X] + lens.size[X]/2 - i, X), s2w(lens.oldPos[Y] + lens.size[Y]/2, Y), labelplane );
//						gl.glDrawPixels(1, lens.size[Y], gl.GL_RGB, gl.GL_FLOAT, ylenspixels[lens.size[X] - i]);
//					}
//				}
//				if(diffPos[X] >= 0){	
//					//System.out.println("x more");				
//					for(int i = 0; i <= Math.abs(diffPos[X]); i++){
//						gl.glRasterPos3d(s2w(lens.oldPos[X] - lens.size[X]/2 + i, X), s2w(lens.oldPos[Y] + lens.size[Y]/2, Y), labelplane );
//						gl.glDrawPixels(1, lens.size[Y], gl.GL_RGB, gl.GL_FLOAT, ylenspixels[i]);
//					}
//				}
//			
//			}
//		
//			// read the pixels in two different ordered arrays 
//			// to make the buffer assgnment easier 
//			// it is not very performant 
//			gl.glReadBuffer(gl.GL_BACK);				
//			for(int i=0; i <= lens.size[Y]; i++){
//				gl.glReadPixels(lens.min[X], winsize[Y] -  lens.max[Y] + i, lens.size[X], 1 , gl.GL_RGB, gl.GL_FLOAT, xlenspixels[i]);
//			}
//			for(int i=0; i <= lens.size[X]; i++){
//				gl.glReadPixels(lens.min[X] + i, winsize[Y] -  lens.max[Y], 1, lens.size[Y], gl.GL_RGB, gl.GL_FLOAT, ylenspixels[i]);
//			}	
//			// decide what to read if the land is outside
//		
//			// do a two lines move of the two cells you picked
//			//expandleaves = -1;
//			//resetGridUniform();
//				
//			twoLinesMove(Y, lens.minc[Y], a, lens.maxc[Y], b );
//			twoLinesMove(X, lens.minc[X], a , lens.maxc[X] , b);
//		
//		
//			// change to a smaller viewport depending on the lens size
//			// do not use a reshape here b/c this changes the winsize
//			int gridMax = gridDepth[0] > gridDepth[1] ? gridDepth[0] : gridDepth[1];
//				
//			//gl.glDrawBuffer(GL_BACK);
//			gl.glViewport(lens.min[X],  winsize[Y] - lens.max[Y], lens.size[X], lens.size[Y]);
//			gl.glMatrixMode(GL_PROJECTION);
//			gl.glLoadIdentity();
//			gl.glOrtho(0.0f, 1.0f, 0.0f, 1.0f, .01f, gridMax + 4.0f);
//			gl.glScalef(1.0f, -1.0f, 1.0f);
//			gl.glTranslatef(0.0f, -1.0f, 0.0f);
//			gl.glMatrixMode(GL_MODELVIEW);
//			gl.glLoadIdentity();
//
//			// start a new Frame to make changes visible
//			startNewFrame();
//			
//			// read what you've done
//			gl.glReadBuffer(gl.GL_BACK);
//		
//			gl.glReadPixels(lens.min[X], winsize[Y] - lens.max[Y], lens.size[X], lens.size[Y], gl.GL_RGB, gl.GL_FLOAT, zoompixels);
//		
//			// unfortunately I have to decide what size to read to get a good working lens
//			// even at the boundaries
//			gl.glReadPixels(lens.min[X], winsize[Y] - lens.max[Y], lens.size[X], lens.size[Y], gl.GL_RGB, gl.GL_FLOAT, zoompixels);
//		
//		
//			// switch back to the original viewport
//			gl.glViewport(0,0, winsize[X], winsize[Y]);
//			gl.glMatrixMode(GL_PROJECTION);
//			gl.glLoadIdentity();
//			gl.glOrtho(0.0f, 1.0f, 0.0f, 1.0f, .01f, gridMax + 4.0f);
//			gl.glScalef(1.0f, -1.0f, 1.0f);
//			gl.glTranslatef(0.0f, -1.0f, 0.0f);
//			gl.glMatrixMode(GL_MODELVIEW);
//			gl.glLoadIdentity();
//		
//			// undo the two lines move using the stored temp variables 
//			// stored in updateLens() 
//		//	System.out.println("Splitpos prev " + lens.temp_min[X]);		
//			twoLinesMove(X, lens.minc[X], lens.temp_min[X], lens.maxc[X], lens.temp_max[X]);
//			twoLinesMove(Y, lens.minc[Y], lens.temp_min[Y], lens.maxc[Y], lens.temp_max[Y]);
//		
//		//	System.out.println("Splitpos real " + lens.minc[X].getSplitPos(X, false));
//		
//			//expandleaves = 1;
//			//resetGridUniform();
//		
//			// 	start a new Frame to make changes visible	
//			startNewFrame();
//			//expandLeaves(1);
//			// draw the lens on the front buffer
//			gl.glDrawBuffer(GL_FRONT);
//			glc.setDoubleBuffered(false);
//			gl.glDepthMask(false);
//			gl.glDepthFunc(GL_ALWAYS);
//			gl.glRasterPos3d(s2w(lens.min[X], X), s2w(lens.max[Y], Y), labelplane );
//			gl.glDrawPixels(lens.size[X], lens.size[Y], gl.GL_RGB, gl.GL_FLOAT, zoompixels);			
//		
//			gl.glFlush();
//			glj.gljFree();
//
// 	
		}


	/**
	 * Helper function for {@link #doDrawFrame doDrawFrame} to move
	 * grid lines during animated transitions. 
	 *
	 * First, all objects that are moved (i.e., GridCells resized) go
	 * through a gradual animated transition. Of course the rest of
	 * the cells are all shrunk or expanded by the same ratio to
	 * accomodate these transitions. 

	 * @author   Tamara Munzner, Serdar Tasiran
	 * @see      AccordionDrawer.GridCell
	 */
	protected void doFrameMove() {
	keepMoving = false;
	
	if (toMove.size() > 0) {
		transitionAnimating = true;
//		requestRedrawEventually();
		Enumeration enum = toMove.elements();
		while (enum.hasMoreElements()) {
		SplitTransition st = (SplitTransition)enum.nextElement();
//		if (st.getDrawer() != this)
//			System.out.println("What???");
		st.incr();
		if (st.done()) {
			toMove.remove(st.getHashKey());
			if (toMove.size() < 1)
			{
				transitionAnimating = false;
				break;
			} 
		} else {
			st.move();
			keepMoving = true;
		}
		}
	}
	}

    public void setDynamicFrameTime(long milliseconds) {dynamicTime = milliseconds;}

    public int getFrameNum() { return frameNum;}
    public Color getBackgroundColor() { return backgroundColor; }
    public Color getLabelColor() { return labelColor; }
    public Color getLabelBackColor() { return labelBackColor; }
    public Color getLabelHiColor() { return labelHiColor; }
    public Color getLabelBackHiColor() { return labelBackHiColor; }
    public Color getRubberbandColor() { return rubberbandColor; }
    public void setBackgroundColor(Color backgroundColor) { this.backgroundColor = backgroundColor;}
    public void setLabelColor(Color labelColor) { this.labelColor = labelColor;}
    public void setLabelBackColor(Color labelBackColor) { this.labelBackColor = labelBackColor;}
    public void setLabelHiColor(Color labelHiColor) { this.labelHiColor = labelHiColor;}
    public void setLabelBackHiColor(Color labelBackHiColor) { this.labelBackHiColor = labelBackHiColor;}
    public void setRubberbandColor(Color rubberbandColor) { this.rubberbandColor = rubberbandColor;}


    public void setNumAnimSteps(int numAnimSteps) { this.numAnimSteps = numAnimSteps; }
    public int getNumAnimSteps() { return numAnimSteps; }

    public void setInflateIncr(double value) { inflateIncr = value;}
    public double getInflateIncr() { return inflateIncr; }

    public void setFlash(CellGeom cg, Color col, int x, int y, boolean doBox) {
//	System.out.println("setflash start" + x + " " + y);
	flashGeom = cg;
	flashCol = col;
	flashX = x;
	flashY = y;
	this.doBox = doBox;
	doFlash();
//	System.out.println("setflash stop" + x + " " + y);
    }


	protected void setHighlightPlane(float w)
	{
	  	hiliteplane = w;		
	}
	protected void resetHighlightPlane()
	{
			hiliteplane = -0.3f;		
	}

	protected void setObjectPlane(float w)
	{
		objplane = w;		
	}
	
	protected void resetObjectPlane()
		{
			objplane = -.5f; 
		}
	
    public void setFocusCell(GridCell fc) {focusCell = fc;}

    public boolean getBigGrid() {return biggrid;}
    public void setBigGrid(boolean on) {biggrid = on; requestRedraw();}
    public void toggleBigGrid() {setBigGrid(!biggrid);}

    public boolean getCacheRange() {return cacherange;}
    public void setCacheRange(boolean on) {
	cacherange = on;
	if (dumpstats) {
	    stats.println();
	    stats.println("CACHERANGE: "+cacherange);
	}
    }
    public void toggleCacheRange() {setCacheRange(!cacherange);}

    public boolean getCheckTime() {return checktime;}
    public void setCheckTime(boolean on) {
	checktime = on; 
	if (dumpstats) {
	    stats.println(); 
	    stats.println("CHECKTIME: "+checktime);
	}
    }
    public void toggleCheckTime() {setCheckTime(!checktime);}

    public boolean getColorGrid() {return colorgrid;}
    public void setColorGrid(boolean on) {colorgrid = on; requestRedraw();}
    public void toggleColorGrid() {setColorGrid(!colorgrid);}

    public boolean getDimBrite() {return dimbrite;}
    public void setDimBrite(boolean on) {dimbrite = on; requestRedraw();}
    public void toggleDimBrite() {setDimBrite(!dimbrite);}

    public boolean getDimColors() {return dimcolors;}
    public void setDimColors(boolean on) {dimcolors = on; requestRedraw();}
    public void toggleDimColors() {setDimColors(!dimcolors);}

    public boolean getDrawLabels() {return drawlabels;}
    public void setDrawLabels(boolean on) {drawlabels = on; requestRedraw();}
    public void toggleDrawLabels() {setDrawLabels(!drawlabels);}

    public int getDrawOrder() {return draworder;}
    public void setDrawOrder(int order) {
	draworder = order;
	if (dumpstats) {
	    stats.println();
	    stats.println("DRAWORDER: "+draworder);
	}
    }
    public void toggleDrawOrder() {
	draworder++;
	if (draworder > FROMFOCUSCELL) draworder = BRUTEFORCE;
	setDrawOrder(draworder);
    }

    public boolean getDumpStats() {return dumpstats;}
    public void setDumpStats(boolean on) {
	dumpstats = on; 
	if (on) {
	    try {
		String fname = "stats"+key+"."+System.currentTimeMillis();
		stats = new PrintWriter(new BufferedWriter(new FileWriter(fname)));
	    } catch (IOException  e) {
		System.out.println("can't open benchmark file: stats");
	    }
	} else {
	    stats.flush();
	    stats.close();
	}
    }
    public void toggleDumpStats() {setDumpStats(!dumpstats);}

    public int getExpandLeaves() {return expandleaves;}
    public void toggleExpandLeaves() {
	expandleaves++;
	if (expandleaves > 1) expandleaves = -1;
	setExpandLeaves(expandleaves, true);
    }
    public void setExpandLeaves(int expand, boolean redraw) {
	expandleaves = expand;
	resetGridUniform();
	if (redraw)
		requestRedraw();
    }


    public boolean getGuarVis() {return guarvis;}
    public void setGuarVis(boolean on) {guarvis = on; requestRedraw();}
    public void toggleGuarVis() {setGuarVis(!guarvis);}

    public boolean getLabelDrawBack() {return labeldrawback;}
    public void setLabelDrawBack(boolean on) {labeldrawback = on; requestRedraw();}
    public void toggleLabelDrawBack() {setLabelDrawBack(!labeldrawback);}

    public boolean getLabelPopup() {return labelpopup;}
    public void setLabelPopup(boolean on) {labelpopup = on;}
    public void toggleLabelPopup() {setLabelPopup(!labelpopup);}

    public boolean getLabelPosRight() {return labelposright;}
    public void setLabelPosRight(boolean on) {labelposright = on; requestRedraw();}
    public void toggleLabelPosRight() {setLabelPosRight(!labelposright);}

    public boolean getLabelTransp() {return labeltransp;}
    public void toggleLabelTransp() {setLabelTransp(!labeltransp);}
    
    public void setLabelTransp(boolean on) {
    	// please don't use bleding for now
    	// could fuck up rendering state under one of the platforms
    }

    public int getLineThickness() {return linethickness;}
    public void setLineThickness(int thickness) {
	linethickness = thickness;
	if (linethickness < 1) linethickness = 1;
//	System.out.println("lt "+linethickness);
	requestRedraw();
    }
    public void increaseLineThickness() {setLineThickness(linethickness+1);}
    public void decreaseLineThickness() {setLineThickness(linethickness-1);}

    public boolean getLinkedNav() {return linkednav;}
    public void setLinkedNav(boolean on) {linkednav = on;}
    public void toggleLinkedNav() {linkednav = !linkednav;}


    public boolean getNoFlash() {return noflash;}
    public void setNoFlash(boolean on) {noflash = on; requestRedraw();}
    public void toggleNoFlash() {setNoFlash(!noflash);}

    public boolean getNoGeoms() {return nogeoms;}
    public void setNoGeoms(boolean on) {nogeoms = on; requestRedraw();}
    public void toggleNoGeoms() {setNoGeoms(!nogeoms);}

    public boolean getNoGrid() {return nogrid;}
    public void setNoGrid(boolean on) {nogrid = on; requestRedraw();}
    public void toggleNoGrid() {setNoGrid(!nogrid);}

    public boolean getShowAlg() {return showalg;}
    public void setShowAlg(boolean on) {
    	boolean originalDoubleBuffer = doublebuffer;
	showalg = on;
//	setDoubleBuffer(!showalg, true, false); // only for a clear??
	requestRedraw();
//	setDoubleBuffer(originalDoubleBuffer, false);
    }
    public void toggleShowAlg() {setShowAlg(!showalg);}

    public void reset() {
//		boolean originalDoubleBuffer = getDoubleBuffer();
//		setDoubleBuffer(doDoubleBuffer, true); // just to clear?  this is eventually done in requestRedraw
	resetGridUniform();
	focusCell = defaultFocusCell;
	keepMoving = false;
	requestRedraw();
//	setDoubleBuffer(originalDoubleBuffer, false);
    }	

    public void changedMarks() {
	changedMarksFrame = frameNum;
    }

    public void clearQueue() {
	ToDrawQ.clear();
    }
    
//    public void addCellToQueue(GridCell cell) { 
//	if (cell == null) return;
//	ToDrawQ.add(cell); // adding uses the comparator function
//	cell.enqueuedFrame = frameNum;
//    }


//    public GridCell findEnclosingCell(int arow, int acol, int brow, int bcol) {
//	int i;
//	int aShift[] = new int[2];
//	int bShift[] = new int[2];
//	aShift[0] = acol;
//	aShift[1] = arow;
//	bShift[0] = bcol;
//	bShift[1] = brow;
//	int depth = 0;
//	while ( (aShift[0] != bShift[0]) || (aShift[1] != bShift[1])) {
//	    for (i = 0; i < 2; i++) {
//		aShift[i] = aShift[i]>>1; 
//		bShift[i] = bShift[i]>>1; 
//	    }
//	    depth++;
//	}
//        return splitTree.getGridCell(depth, aShift[1], aShift[0]);
//    }



	public void resetGridUniform() {
	for (int xy = 0; xy < 2; xy++) {
		splitLine[xy].resetSplitValues();
	}

	toMove.clear(); // stop all animations
	transitionAnimating = false;
	}

    /**
    * Converts pixel coordinates into relative "fractional" coordinates in [0,1]
    */ 	
    public double s2w(int s, int xy) { return ((double)s)/winsize[xy];}
    public double s2w(double s, int xy) { return ((double)s)/winsize[xy];}

    /**
    * Converts "fractional" coordinates in [0,1] into pixel coordinates
    */ 
    public int w2s(double w, int xy) { return (int) Math.floor(w*winsize[xy]);}

    public int getWinMaxX() { return winsize[X]; }
    public int getWinMaxY() { return winsize[Y]; }
    public int getWinMax(int xy) { return winsize[xy]; }

    public abstract CellGeom pickGeom(int x, int y);
//	return rootCell.pickDescend(x, y);
//    }

    public void setCellHighlight(GridCell c, boolean on) {
	c.setDrawBackground(on);
    }

    public void setColorGL(Color col) {
		float thecol[] = new float[3];
		col.getRGBColorComponents(thecol);
		gl.glColor3f(thecol[0], thecol[1], thecol[2]);
    } 
    public boolean drawStart(boolean swap) { 
//System.out.println("drawStart " + swap);
    	boolean returnValue = glj.gljMakeCurrent();
    	if (swap) glj.gljSwap();
    	return returnValue;
	}
    
    public void drawEnd(boolean swap) { 
    	//System.out.println("drawEnd " + swap);
	gl.glFlush();
	if (swap) glj.gljSwap();
//	glj.gljCheckGL(); // probably not necessary?
	glj.gljFree();
    }

	


    public abstract CellGeom pickAttached(int x, int y); // ArrayList of CellGeoms
    public abstract void initCells();
    public abstract ArrayList getColorsForGridCell(GridCell c, int objkey);
    public abstract ArrayList getColorsForCellGeom(CellGeom c);
    public abstract void doFlash();
    public abstract void drawNextFancy(CellGeom g);
    abstract protected void drawPreContFrame();
    abstract protected void drawPreNewFrame();
    abstract protected void drawPostFrame();
    abstract protected void drawPostScene();
    abstract protected void drawBruteForce();
    abstract protected void seedQueue();
    abstract public void subpixelDraw(GridCell c);
    public abstract OlduvaiObject getOlduvaiObject();
    public abstract void drawReferenceLine();
    OlduvaiObject o;
    
    abstract public int getBottomGrid( boolean horizontal );

	/**
	 * @return
	 */
	public double getMinCellDims(int xy) {
		return minCellDims[xy];
	}

	/**
	 * @param is
	 */
	public void setMinCellDims(int xy, double is) {
		minCellDims[xy] = is;
	}

	/**
	 * @return
	 */
	public double getObjplane() {
		return objplane;
	}

	/**
	 * @return
	 */
	public double getLabelplane() {
		return labelplane;
	}

	/**
	 * @return
	 */
	public boolean isDoingFlash() {
		return doingFlash;
	}

	/**
	 * @return
	 */
	public int getWinsize(int xy) {
		return winsize[xy];
	}
	
	public void setDrawSplits(boolean on)
	{
		drawsplits = on;
	}
	
	public void toggleDrawSplits()
	{
		drawsplits = !drawsplits;
	}

	/**
	 * @return
	 */
	public ArrayList getDrawnLabels() {
		return drawnLabels;
	}

	protected static final int FAR_FROM_STUCK = -1;
	protected static final int MIN_STUCK_X = 0;
	protected static final int MIN_STUCK_Y = 1;
	protected static final int MAX_STUCK_X = 2;
	protected static final int MAX_STUCK_Y = 3;
	// -1: not close to a stuck line
	//  0: close to min stuck X
	//  1: close to min stuck Y
	//  2: close to max stuck X
	//  3: close to max stuck Y
	// create an interaction box on a stuck line to mark that a stuck line has been
	//  chosen to move with the next drag
	protected InteractionBox createBoxFromCells(int stuckType)
	{
//		System.out.println("Create box from cells with stuckType: " + stuckType);
		if (stuckType == FAR_FROM_STUCK) return null;
		int minSplit[] = {-1, -1};
		int maxSplit[] = {splitLine[X].size, splitLine[Y].size};
		if (stuckType == MIN_STUCK_X)
			maxSplit[X] = -1;
		else if (stuckType == MIN_STUCK_Y)
			maxSplit[Y] = -1;
		else if (stuckType == MAX_STUCK_X)
			minSplit[X] = splitLine[X].size;
		else if (stuckType == MAX_STUCK_Y)
			minSplit[Y] = splitLine[Y].size;
		return new InteractionBox(minSplit, maxSplit, null, this);
	}

	/**
		 * Make an InteractionBox from the GridCells nearest to the box of the mousedrag.
		 * Expects dragStart[X]/Y, dragEnd[X]/Y to be set. 
		 * 
		 * @author Francois Guimbretiere, Tamara Munzner */
	protected InteractionBox createBoxFromCells(int[] dragStart, int[] dragEnd) {
		// Convert from the start and end of a drag motion to absolute
		// min/max coordinates so that computation can be carried out
		// independent of the direction that the user chose to drag.
		int min[] = new int[2];
		int max[] = new int[2];
		for (int xy = 0; xy < 2; xy++) {
			if (dragStart[xy] < dragEnd[xy]) {
				min[xy] = dragStart[xy];
				max[xy] = dragEnd[xy];
			} else {
				min[xy] = dragEnd[xy];
				max[xy] = dragStart[xy];
			}
		}
		int[] minSplit = {splitLine[X].getMinIndexForPixelValue(min[X], frameNum), 
				splitLine[Y].getMinIndexForPixelValue(min[Y], frameNum)};
		int[] maxSplit = {splitLine[X].getMinIndexForPixelValue(max[X], frameNum)+1,
				splitLine[Y].getMinIndexForPixelValue(max[Y], frameNum)+1};
		return new InteractionBox(minSplit, maxSplit, null, this);//)minCell, maxCell, null, this));
	
	}
	/**
		 * Make an InteractionBox from the GridCells nearest to the box of the mousedrag.
		 * Expects dragStart[X]/Y, dragEnd[X]/Y to be set. 
		 * 
		 * @author Francois Guimbretiere, Tamara Munzner */
	// baseBox.dragStart, dragEnd positions are in pixels; where mouse started and is now while dragging
	// basebox is defined by 4 split lines (2 movable, 2 not) and is currently being resized
	
	// CRFP: return 2 integers, pixel differences for baseBox movable lines in X and Y
	protected int[] createRectFromPick(InteractionBox baseBox) {
		// if they pick close enough to a corner, 
		// use baseBox for the other corner
//		System.out.println("interactions happen here, start to log them or something");

		int[] returnArray = new int[2]; // x min, y min, x max, y max (-1 for unchanged values)
		for (int xy = X; xy <= Y; xy++)
		{
			int pixelMinContextInside = w2s(minContextInside, xy);
			int stuckPixelPosition = w2s(splitLine[xy].getAbsoluteValue(baseBox.stuckLineIndex[xy], frameNum), xy);
			int dragLength = baseBox.dragStart[xy] - baseBox.dragEnd[xy];
			returnArray[xy] = baseBox.originalMovePixelPosition[xy] - dragLength;
			if (baseBox.moveLineIndex[xy] > baseBox.stuckLineIndex[xy] &&
				returnArray[xy] < stuckPixelPosition + pixelMinContextInside)
			{ // moveLine is bigger than stuck, max line is moving	
				returnArray[xy] = stuckPixelPosition + pixelMinContextInside;
			}
			else if (baseBox.moveLineIndex[xy] < baseBox.stuckLineIndex[xy] &&
				returnArray[xy] > stuckPixelPosition - pixelMinContextInside)
			{ // moveline is smaller than stuck, min line is moving
				returnArray[xy] = stuckPixelPosition - pixelMinContextInside;
			}
		}
		return returnArray;
	}

	// assert baseBox is not null
	protected void moveStuckPosition(InteractionBox baseBox)
	{
		while (transitionAnimating)
			endAllTransitions(); // end transitions

		// CRFP: return 2 integers, pixel differences for baseBox movable lines in X and Y
		int[] rectBox = createRectFromPick(baseBox);
		double newPos;
		int xy = X;
		if (baseBox.getMinLine(X) != baseBox.getMaxLine(X))
			xy = Y;
		double stuckValue[] = {splitLine[xy].getMinStuckValue(), splitLine[xy].getMaxStuckValue()};
		
		if (baseBox.getMinLine(xy) == -1) // minstuck moving
		{
			newPos = s2w(rectBox[xy], xy);
			if (newPos > stuckValue[1] - minContextInside)
				splitLine[xy].setMinStuckValue(stuckValue[1] - minContextInside);
			else if (newPos < SplitLine.defaultMinStuckValue)
				splitLine[xy].setMinStuckValue(SplitLine.defaultMinStuckValue); 
			else
				splitLine[xy].setMinStuckValue(newPos);
		}
		else // maxStuck moving
		{
			newPos = s2w(rectBox[xy], xy);
			if (newPos < stuckValue[0] + minContextInside)
				splitLine[xy].setMaxStuckValue(stuckValue[0] + minContextInside);
			else if (newPos > SplitLine.defaultMaxStuckValue)
				splitLine[xy].setMaxStuckValue(SplitLine.defaultMaxStuckValue);
			else
				splitLine[xy].setMaxStuckValue(newPos);
		}
//		System.out.println("newPos: " + newPos + " oldStuck: " + stuckValue[0] + " " + stuckValue[1]);
			
		incrementFrameNumber(); // so compute place this frame will update, this can be moved to after transitions are created
		keepMoving = true; // a user movement?
		requestRedraw(); // redraw the frame, this will enact the movements in the move queue
	}


	/**
	 * @return
	 */
	public Vector getToDrawQ() {
		return ToDrawQ;
	}
	
	/**
	 * @return
	 */
	public SplitLine getSplitLine(int xy) {
		return splitLine[xy];
	}
	
	public void incrementFrameNumber()
	{
		frameNum++;
	}
	
	/**
	 * @return
	 */
	public InteractionBox getFlashBox()
	{
		return flashBox;
	}

	/**
	 * @return Returns the flashCol.
	 */
	public Color getFlashCol() {
		return flashCol;
	}
	
	public void drawText(double x, double y, String name, int fontheight, Color col, double zplane,
			Color outlineColor)
	{
//		if (outlineColor != null)
//		{
//			final int shadowDepth = 1;
//			double delta[] = { s2w(shadowDepth, X), s2w(shadowDepth, Y)};
//			for (int xc = -1; xc < 2; xc+=2)
//				for (int yc = -1; yc < 2; yc+=2)
//					drawText(x + xc * delta[X], y + yc * delta[Y], name, fontheight, outlineColor, zplane - 0.01, null);			
//		}
		if (takeSnapshot)
		{
			try
			{
				snapShotWriter.write("/Arial findfont " +
					fontheight + " scalefont setfont " +
					w2s(x, X) + " " + w2s(y, Y) + " moveto " +
					"gsave 1 -1 scale (" + name + ") " +
					col.getRed() / 255f	+ " " + 
					col.getGreen() / 255f + " " + 
					col.getBlue() / 255f + " setrgbcolor show grestore\n");
			} catch (IOException ioe) {
				System.out.println("Error: IOException while trying to write cell names to file: "
					+ snapShotWriter.toString());
			}
		}
		else
		{
			if (outlineColor != null)
			{
				final int shadowDepth = 1;
				double delta[] = { s2w(shadowDepth, X), s2w(shadowDepth, Y)};
				for (int xc = -1; xc < 2; xc+=2)
					for (int yc = -1; yc < 2; yc+=2)
						drawText(x + xc * delta[X], y + yc * delta[Y], name, fontheight, outlineColor, zplane - 0.01, null);			
			}
			setColorGL(col);
			gl.glRasterPos3d(x, y, zplane);
			bff.drawString(gl, name, fontheight);
		}
	}
	
}
