
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionSequenceDrawer;

import java.util.*;
import java.awt.*;
import java.io.*;

import gl4java.*;

import AccordionDrawer.*;

/**
 * An abstract class representing a multi-scale rectangular grid on which 
 * a tree consisting of horizontal and vertical line segments is drawn
 *
 * Fills in the fields of the abstract class AccordionDrawer based on
 * information from the Tree class.
 * 
 * @author  Tamara Munzner, Li Zhang
 * @version 
 * @see     AccordionDrawer.AccordionDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.TreeEdge
 * @see     AccordionDrawer.GridCell
 */

//public class AccordionTreeDrawer extends AccordionDrawer {
public abstract class AccordionSequenceDrawer extends AccordionDrawer {
	String fname;
	//protected double pointoffset[] = new double[2];
	protected int labeloffset[] = new int[2];
	protected int labelbuffer[] = new int[2];

	/**
	 * framebuffer pixels from last mouseover label drawn
	 */
	protected float pixels[];
	
	/**
	 * GL rasterpos of last mouseover label drawn
	 */
	protected int drawnlabelpos[] = new int[2];
	/** 
	 * width and height of last mouseover label drawn
	 */
	protected int drawnlabelsize[] = new int[2];
	boolean biggerThanScreen = false;
	
	//how the labels are drawn
	// dealing with variable font size
	public int minFontHeight = 14;
	public int maxFontHeight = 14;
	public int popupFontHeight = 14;
	
	public static boolean complementFlag = false;
	
	// use this to locate next non gap in a range
	protected static RangeList gapRangeList = null; // set this to the gap ranges
	
	
	protected double forestShrinkFactor;
	protected double pickoffset[] = new double[2];
	protected int pickFuzz;
	
	
	

	public static boolean fullyQualified;

	// benchmarking 
	public static int countDrawnFrame;
	public static int countDrawnScene;
	public static int cumFrameTime;
	
	public int[] defaultFocusRange = new int[2];

	public ArrayList sequences;

	public AccordionSequenceDrawer(ArrayList seq, int w, int h) {
		super(w, h);
		expandleaves = 0;
//		BinaryTree binaryTree = (BinaryTree)splitTree;
		horiz = true; // draw sequences horizontally

		int namewidth;

		sequences = seq;
//		for (int i = 0; i < seq.size(); i++)
//		{
//			namewidth = bff.stringWidth(getSequence(i).getName(), getMaxFontHeight());
//			if (namewidth > nameStuckVal) nameStuckVal = namewidth;
//		}
		labeloffset[X] = 2;
		labeloffset[Y] = 4;
		labelbuffer[X] = 10;
		labelbuffer[Y] = 1;

		pickoffset[0] = .1f;
		pickoffset[1] = .1f;
		pickFuzz = 5; // in pixels
		forestShrinkFactor = -0.9f; // -0.3;
		
		
		initCells();
		splitLine[Y].setMinStuckValue(0.021f);
		if (sequences.size() == 1)
			splitLine[X].setMinStuckValue(0.001f);
		else
			splitLine[X].setMinStuckValue(0.3f);

		flashGeomOld = null;
		doBox = false;

		defaultFocusRange[0] = 0;
		defaultFocusRange[1] = 0;
	}

	public boolean drawSpecial()
	{
		drawFrame();
		return true;
	}
	
	public abstract void drawGroups();
	
	public int drawCount;
	static long startRender;
	
	public static int dequeueChunkPerTimeCheck = 20;
	public void drawFrame()
	{
		doFrameMove();
		int xy = splitLine[X].isHorizontal() ? X : Y;
		
	    // Helpful for debugging queue
		if (startFrame)
		{
			
			drawCount = 0;
//			System.out.println("Draw queue size: " + ToDrawQ.size());
			startRender = System.currentTimeMillis();
			drawNameLabels();
			if (!takeSnapshot)
				drawGroups();
//	        System.out.println("After group draw: " + drawCount + " nodes drawn in " + 
//	                (System.currentTimeMillis() - startRender) + " time");
		}

		now = ignoreProgressive ? 0 : System.currentTimeMillis();
		while ((ignoreProgressive || now - dynamicStart < dynamicTime) && ToDrawQ.size() > 0)
		{
			for (int i = 0; ToDrawQ.size() > 0 && i < dequeueChunkPerTimeCheck; i++)
			{
				RangeInSequence r = (RangeInSequence)ToDrawQ.get(0);
				drawRange(r.min-1, r.max+1);
				drawDensity(r.min-1, r.max+1);
				ToDrawQ.remove(0);
			}
			now = ignoreProgressive ? 0 : System.currentTimeMillis();
		}

		// Queue debugging
//	    if (ToDrawQ.size() >= 0)
//	    {
//	        System.out.println(drawCount + " nodes drawn in " + 
//	                (System.currentTimeMillis() - startRender) + " time");
//			System.out.println("Draw queue size: " + ToDrawQ.size());
//	    }
		//System.out.println("After drawing: leaves: " + leafDrawCount + "\tinternals: " + internalDrawCount);
		
		drawPostFrame();
		if (keepMoving)
		{
			keepDrawing = false;
			incrementFrameNumber();
			requestRedrawEventually();
		}
		else if (ToDrawQ.size() > 0)
		{
//			System.out.println(" more to draw " + ToDrawQ.size());
			keepDrawing = true;
			incrementFrameNumber();
			requestRedrawEventually();
		}
		else 
		{
			if (takeSnapshot)
				drawGroups();
	        System.out.println(drawCount + " nodes drawn in " + 
            (System.currentTimeMillis() - startRender) + " time");
			

//			System.out.println(" nothing more to draw ");
			keepDrawing = false;
			if (colorgrid || drawBackground) splitLine[X].drawBackground(null, frameNum);
			drawPostScene();
		}
			
	
	}
	public abstract ArrayList getColorsForRange( int minO, int maxO, int seqMin, int seqMax);
	
	/**
	 * 
	 * @param seqs min and max sequence number
	 * @param sites min and max site number
	 * @param plane z height (usually objplane)
	 */
	public void drawInCells(Color col, String label, int[] seqs, int[] sites, double[] drawingRange, double plane, boolean gv) {
		drawCount++;
		AccordionSequenceDrawer.countDrawnFrame++;
		GridCell.incrementCountBBoxDraw();
//		System.out.println("Drawing at depth: " + plane + ", color: " + col);
		if (nogeoms) return;
		
		final int X = AccordionDrawer.X;
		final int Y = AccordionDrawer.Y;
		
		
		
		Sequence topSeq = getSequence(seqs[0]);
		SiteNode topLeftNode = topSeq.getNode(sites[0]);
		Sequence botSeq = getSequence(seqs[1]);
		SiteNode botRightNode = botSeq.getNode(sites[1]);
		if (topLeftNode == null || botRightNode == null)
		{
			System.out.println("Error: these nodes should not be null: " + 
					topLeftNode + "@" + seqs[0] + "|" + sites[0] + ", " +
					botRightNode  + "@" + seqs[1] + "|" + sites[1]);
			return;
		}
		
		double cellMin[] = {topLeftNode.getMin(X), topLeftNode.getMin(Y)};
		double cellMax[] = {botRightNode.getMax(X), botRightNode.getMax(Y)};
		double cellHeightRange = cellMax[Y] - cellMin[Y];
		cellMax[Y] = cellMin[Y] + drawingRange[1] * cellHeightRange;
		cellMin[Y] = cellMin[Y] + drawingRange[0] * cellHeightRange;
		
		if (gv) // guarantee cells to be a certain dimension, marked cells only
		{
			for (int xy = X; xy <= Y; xy++)
			{
				double minCellDims = getMinCellDims(xy);
//				if (minCellDims < 1)
//					System.out.println("debug");
				if (minCellDims < 3) minCellDims = 3; // guarantee at least 3 pixels if gv is on
				if (false) // this makes overlaps if the marked cell is too small and minCellDims > 1
				{
					if (w2s(cellMax[xy] - cellMin[xy], xy) < minCellDims)
						cellMax[xy] = cellMin[xy] + s2w(minCellDims, xy);
				}
				else // still overlaps if cell is smaller than 1 pixel, but that's what we should minimally guarantee for marks
					if (w2s(cellMax[xy] - cellMin[xy], xy) < 3)
					{
						cellMin[xy] -= s2w(1, xy);
						cellMax[xy] = cellMin[xy] + s2w(3, xy);
					}
			}
		}
		if (takeSnapshot)
		{
			int cellMinPix[] = {w2s(cellMin[X], X), w2s(cellMin[Y], Y)};
			int cellMaxPix[] = {w2s(cellMax[X], X), w2s(cellMax[Y], Y)};
			try
			{
				snapShotWriter.write("newpath " +
				cellMinPix[X] + " " + cellMinPix[Y] + " moveto " +
				cellMinPix[X] + " " + cellMaxPix[Y] + " lineto " +
				cellMaxPix[X] + " " + cellMaxPix[Y] + " lineto " +
				cellMaxPix[X] + " " + cellMinPix[Y] + " lineto " +
				"closepath " +
				"gsave " + col.getRed() / 255f + " " + col.getGreen() / 255f + " " + col.getBlue() / 255f + " setrgbcolor " +
				"eofill grestore\n");
			}
			catch (IOException ioe)
			{
				System.out.println("Error: IOException while trying to write cells to file: " + snapShotWriter.toString());
			}
		}
		else
		{
			setColorGL(col);
			GLFunc gl = (GLFunc) getGraphicsStuff();
			
			gl.glBegin(GLEnum.GL_QUADS);
			gl.glVertex3d(cellMin[X], cellMin[Y], plane);
			gl.glVertex3d(cellMin[X], cellMax[Y], plane);
			gl.glVertex3d(cellMax[X], cellMax[Y], plane);
			gl.glVertex3d(cellMax[X], cellMin[Y], plane);
			gl.glEnd();
		}
		
		
		if (drawlabels && label.length() > 0 && !SiteNode.isGap(label.charAt(0))) {
			drawLabelBox( label, maxFontHeight, cellMin, cellMax, true);
		}
	}
	
	public void drawRange(int min, int max)
	{
		//Vector listOfSequences = null;
		// for each sequence, draw cells between min and max
		final double[] drawingRange = {0.00, 1.0};
		final double[] drawingRangeFull = {0.00, 1.0};
		
		if (min + 1 == max)
		{
			Iterator iter = listOfSequences.iterator();
			int prevMaxSeq = -1;
			while (iter.hasNext())
			{
				int currMaxSeq = ((SplitCell)iter.next()).max;
				int seqs[] = {prevMaxSeq+1, currMaxSeq};
				// find a node, if not all nodes are gaps
				SiteNode n = null;
				int nodeFindSeq = prevMaxSeq + 1;
				while ((n == null || n.isGap()) && nodeFindSeq <= currMaxSeq)
				{
//					System.out.println("get node in seq: " + nodeFindSeq);
					n = getSequence(nodeFindSeq++).getNode(max);
				}
//				System.out.println("Draw in cell: site: " + n);
				if (n != null)
					n.drawInCell(null, objplane, drawingRange, seqs, false);
				else
					System.out.println("Tried to find a non-gap, but settling since no colored sitenode found in " + min + "->" + max + " seqs " + seqs[0] + "->" + seqs[1]);
				prevMaxSeq = currMaxSeq;
			}
			return;
		}
		SplitCell[] sc = {SplitCell.getOverlapSplitCell(listOfSites, new SplitCell(min+1)),
				SplitCell.getOverlapSplitCell(listOfSites, new SplitCell(max))};
		// get the split cell for this range, if one exists
		//  (it should always exist since single cell case is taken care of before reaching here)
		if (sc[0] != null && sc[1] != null && sc[0] == sc[1])
		{
			if (sc[0].cullingObject == null || sc[0].cullingObject instanceof ColumnNumberList)
				new ColumnColorList(sc[0], splitLine[X], sequences); // sc[0].cullobject gets set internally
			TreeSet siteNodeColors = (TreeSet)((ColumnColorList)sc[0].cullingObject).columnColors;
			if (siteNodeColors == null)
			{
				System.out.println("Error: sitenode colors is null for a column color list");
				return;
			}
			Iterator iter = listOfSequences.iterator();
			int prevMaxSeq = -1;
			int[] sites = {min+1, max};
			
			while (iter.hasNext())
			{
				
				int currMaxSeq = ((SplitCell)iter.next()).max;
				while (currMaxSeq <= prevMaxSeq && iter.hasNext())
					currMaxSeq = ((SplitCell)iter.next()).max;
				int[] seqs = {prevMaxSeq+1, currMaxSeq};
				
				ColumnColorSet.getColor = true;
				ColumnColorSet.returnColorSet = null; // debug: see if it's changed
				if (siteNodeColors.contains(new ColumnColorSet('-', seqs[0], seqs[0])))
				{
					ColumnColorSet ccs = ColumnColorSet.returnColorSet;
					if (SiteNode.isGap(ccs.colorLetter) && ccs.maxSeq < seqs[1])
					{ // try to find something that is not a gap
						ColumnColorSet.getColor = true;
						int nextMin = ccs.maxSeq + 1;
						if (siteNodeColors.contains(new ColumnColorSet('-', nextMin, nextMin)))
						{
							ccs = ColumnColorSet.returnColorSet;
							if (SiteNode.isGap(ccs.colorLetter))
								System.out.println("Tried to find a non-gap, but settling since no colored sitenode found in " + min + "->" + max + " seqs " + seqs[0] + "->" + seqs[1]);
//							else
//								System.out.println("This isn't a gap, but we started with one here: " + min + "->" + max + " seqs " + seqs[0] + "->" + seqs[1]);
							//System.out.println("Draw in cells: sites: " + sites[0] + " -> " + sites[1] + " seqs: " +seqs[0] + " -> " + seqs[1]);
							drawInCells(Sequence.getColor(ccs.colorLetter), ccs.colorLetter+"", seqs, sites, drawingRangeFull, objplane, false);
						}
						else
							System.out.println("siteNodeColors does not contain the color set for " + nextMin);
					}
					else
					{
//						System.out.println("Found a sitenode that isn't a gap right away here: " + min + "->" + max + " seqs " + seqs[0] + "->" + seqs[1]);
						drawInCells(Sequence.getColor(ccs.colorLetter), ccs.colorLetter+"", seqs, sites, drawingRangeFull, objplane, false);
					}
				}
				else
					System.out.println("siteNodeColors does not contain the color set for " + seqs[0]);
				prevMaxSeq = currMaxSeq;
			}
		}
		else
		{
			System.out.println("Couldn't find a split cell for this range: " + min + " " + max);
		}
	}
	
	// build all splitLines
	private void buildCells() {

		int numrows = sequences.size();
		int numcols = getSequence(0).getNumOfNodes();
		// numcols same for all seqs, padded.

		for (int b = 0; b < numrows; b++) { // horizontal binary tree init
			Sequence currSeq = ((Sequence) sequences.get(b));
			// go over each site node in the sequence
			for (int a = 0; a < numcols; a++) { // place all the leaves
				char n = currSeq.getNodeName(a);
				if (!SiteNode.isGap(n)) {
					GridCell c = new GridCell(this);
					splitLine[X].addCell(c, a);
					splitLine[Y].addCell(c, b); // create vertical split for vertical binary tree
				}
			}
			
//			// reverse lookup for gaps; this will be removed when all gaps are in a group
//			GridCell lastGridCellSeen = null;
//			for (int a = currSeq.getNumOfNodes()-1; a >=0; a--)
//			{
//				SiteNode n = currSeq.getNode(a);
//				if (!n.isGap())
//					lastGridCellSeen = n.getCell();
//				else
//					n.setCell(lastGridCellSeen); // for fast lookup of gap distance given first sitenode
//			}
		}
	}





	/**
	 * Initialization done after graphics context has been established.
	 * @author Tamara Munzner
	 */
	public void init() {
		super.init();
		reshape(winsize[X], winsize[Y]);
	}

	/**
	 * Initializes level 0 GridCells based on Tree tree, assign tree
	 * nodes to them. Set {@link #maxUsed maxUsed[]} in directions X
	 * and Y.
	 *
	 *
	 * @author   Tamara Munzner, Serdar Tasiran
	 * 
	 *
	 * @see      AccordionDrawer.GridCell
	 * @see      AccordionDrawer.Tree
	 * @see AccordionDrawer.TreeNode */

	public void initCells() {
	    
		int longestSoFar = 0;
		Iterator iter = sequences.iterator();
		while (iter.hasNext()) {
		Sequence s = (Sequence) iter.next();
		longestSoFar = (s.getNumOfNodes() > longestSoFar) ? s.getNumOfNodes() : longestSoFar;
		}
		bottomSize[0] = longestSoFar; 
		bottomSize[1] = sequences.size();
		initSplitLines();
		gridDepth = new int[2];
		double logBase2 = Math.log(2);
		for (int xy = 0; xy < 2; xy++)
		{
			gridDepth[xy] = (int)Math.ceil(Math.log(bottomSize[xy]) / logBase2);
			gridDepth[xy] = gridDepth[xy] <= 0 ? 1 : gridDepth[xy];
		}
		System.out.println("gridDepth: "+gridDepth[0] + " " +gridDepth[1] +" rows: bottomSize0 "+bottomSize[0]+" cols bottomSize1 "+bottomSize[1]);
		buildCells();
	}

	// x and y are pixel positions
	public CellGeom pickAttached(int x,	int y)
	{
		return pickGeom(x, y);
	}
	
	// pick the sequence at pixel position y
	private Sequence pickSequence(int y, double pixelSize)
	{
		double position = s2w(y, Y);
		int seqPosition = splitLine[Y].getSplitLineIndex(position, pixelSize, frameNum);
		if (seqPosition < 0 || seqPosition > sequences.size() - 1)
			return null;
		SplitCell sc = new SplitCell(seqPosition);
		SplitCell foundSeq = SplitCell.getOverlapSplitCell(listOfSequences, sc);
		return getSequence(foundSeq.max);
	}
	
	// given some pixel position, return the gridCell
	public CellGeom pickGeom(int x, int y)
	{
		if (listOfSequences == null || listOfSites == null)
			return null;
		Sequence seq = pickSequence(y, s2w(getMinCellDims(Y), Y));
		if (seq == null) return null;
		SiteNode n = seq.pickSiteNode(s2w(x, X), splitLine[X], s2w(getMinCellDims(X), X));
		if (n == null) return null;
		return n;
	}

	public void addRangeToDrawQueue(RangeInSequence r)
	{
		if (r != null)
			ToDrawQ.add(r);
	}

	/** 
	 * Create box enclosing the subtree beneath input node/edge. The
	 * resulting box is the nearest GridLine boundaries to that point,
	 * which are cached so that redraws to erase the box can happen
	 * after the lines change position.
	 * @param cg TreeNode to act upon, passed in as CellGeom from
	 *  AccordionDrawer (where TreeNodes are not known)
	 * @see IntersectionBox
	 * @author Tamara Munzner */
	public InteractionBox makeBox(CellGeom cg) {
		if (null == cg)
			return null;
		SiteNode sn = (SiteNode) cg;
		int[] minSplits = new int[2];
		int[] maxSplits = new int[2];
			
		SplitCell foundSiteSC = SplitCell.getOverlapSplitCell(listOfSites, new SplitCell(sn.position));
		if (foundSiteSC == null)
		{
			minSplits[X] = sn.position-1;
			maxSplits[X] = sn.position;
		}
		else
		{
			minSplits[X] = foundSiteSC.min;
			maxSplits[X] = foundSiteSC.max;
		}

			SplitCell foundSeqSC = SplitCell.getOverlapSplitCell(listOfSequences, new SplitCell(sn.seq.key));
			if (foundSeqSC == null)
			{
				minSplits[Y] = sn.seq.key-1;
				maxSplits[Y] = sn.seq.key;
			}
			else
			{
				minSplits[Y] = foundSeqSC.min;
				maxSplits[Y] = foundSeqSC.max;
			}
		return new InteractionBox(minSplits, maxSplits, sn, this);
	}

	
	public abstract double[] getGroupDrawingRange(int groupNum);

	/**
	 * On mouseover, flash the object under the mouse and draw its
	 * label at maximum size.
	 *
	 * @author   Tamara Munzner
	 * 
	 *
	 * @see      AccordionDrawer.GridCell
	 * @see      AccordionDrawer.Tree
	 * @see      AccordionDrawer.TreeNode
	 */
	// represent flash over a range of cells
	// - remove flash geoms, keep flash boxes
	
	// don't flash until view has finished drawing: 
	//  a block that hasn't been rendered will be saved as the background color
	//   if that region hasn't been drawn, then next move will overwrite the new color with background color

	// entering in to this function, depth tests are off and stay off (i.e. don't change functions in AccordionDrawer called from here)
	public void doFlash() {
//		System.out.println("doFlash start");
		final double[] nodeDrawingRange = {0.0, 1.0};
		
		if (noflash
			|| null == glj
			|| keepDrawing // wait until drawing has stopped before turning on flashing
			|| !drawStart(false)) // no swap
			return;
		boolean doubleBufferedOriginal = getDoubleBuffer();
		setDoubleBuffer(false, false); // deactivate double buffering for drawing flash, redrawing old flash
		doingFlash = true;
		
//		 1 pixel around the flash box, off-by-one differences between drawing position and gl functions fix
		final int flashBoxBuffer = 1; 

		if (doBox && flashGeom != null) {
			flashBox = makeBox(flashGeom);
		}
		if (doBox && flashBoxOld != null){
			flashBoxOld.undraw();
		}

		if (flashGeom != null && flashBoxOld != null)
		{
			// draw in old flash cell
			double x= s2w(flashBoxOld.getMin(X) - flashBoxBuffer,X);
			double y = s2w(flashBoxOld.getMax(Y) + flashBoxBuffer,Y);
			if (x < 0)
				x = 0;
			
			gl.glRasterPos3d(x,y, getLabelplane());
			gl.glDrawPixels( drawnlabelsize[0], drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, pixels);
		}
		
		if (null != flashGeom) {
			// read old cells, draw in new flash cell
			// readpixels first
			drawnlabelpos[X] = flashBox.getMin(X) - flashBoxBuffer;
			if (drawnlabelpos[X] < 0)
				drawnlabelpos[X] = 0;
			drawnlabelsize[X] = flashBox.getMax(X) - drawnlabelpos[X] + flashBoxBuffer;
			

			drawnlabelsize[Y] = flashBox.getMax(Y) - flashBox.getMin(Y) + 2 * flashBoxBuffer;
			drawnlabelpos[Y] = flashBox.getMax(Y) + flashBoxBuffer;

			pixels = new float[drawnlabelsize[X] * drawnlabelsize[Y] * 3];

			// read pixels from frontbuffer
			gl.glReadBuffer(GLEnum.GL_FRONT);
			gl.glReadPixels(drawnlabelpos[X], getWinsize(Y) - drawnlabelpos[Y],
					drawnlabelsize[X], drawnlabelsize[Y], GLEnum.GL_RGB,
					GLEnum.GL_FLOAT, pixels);

			// draw flash
			int[] seqs = { flashBox.getMinLine(Y) + 1, flashBox.getMaxLine(Y) };
			int[] sites = { flashBox.getMinLine(X) + 1, flashBox.getMaxLine(X) };
			drawInCells(flashCol, "", seqs, sites, nodeDrawingRange, interactionplane, false);

		}
		if (doBox && flashBox != null)
		{
			flashBox.draw(rubberbandColor, flashBoxWidth, interactionplane);
		}
		
		if (doBox)
			flashBoxOld = flashBox;
		flashGeomOld = flashGeom;
		flashXOld = flashX;
		flashYOld = flashY;
		
		doingFlash = false;
		setDoubleBuffer(doubleBufferedOriginal, false);
		drawEnd(false); // no swap
		
		
	}

	/** 
	 * Decide what GridCells to enqueue for drawing based on tree
	 * topology. We want to add parents and children of current
	 * CellGeom. We should find the GridCell corresponding to the
	 * higher-level TreeEdge associated with a TreeNode, not the
	 * level0 GridCell of the TreeNode itself.
	 * 
	 * @author Tamara Munzner
	 */

	public void drawNextFancy(CellGeom g) {
	    return;
	}

	public TreeSet listOfSequences = null;
	public TreeSet listOfSites = null;
	

	
	public int getBlocksIn(int xy)
	{
		if (xy == X)
			return listOfSites.size();
		else
			return listOfSequences.size();
	}
	
	protected void drawPreNewFrame() {
		countDrawnFrame = 0;
		countDrawnScene = 0;
		cumFrameTime = 0;
		
		listOfSequences = splitLine[Y].makePixelRanges(frameNum);
		listOfSites = splitLine[X].makePixelRanges(frameNum);
		
		gl.glDepthFunc(GLEnum.GL_LEQUAL);
		gl.glEnable(GLEnum.GL_DEPTH_TEST);
		
		int xSize = winsize[X];
		int ySize = winsize[Y];
		if (takeSnapshot)
		{
			try
			{
				snapShotWriter.write("%!PS-Adobe-2.0 EPSF-2.0\n" +
				"%%BoundingBox: 0 0 " + xSize + " " + ySize + "\n" +
				"%%Magnification: 1.0000\n" +
				"%%EndComments\n" +
				
				"gsave\n" +
				
				"newpath " +
				"0 " + ySize + " moveto " +
				"0 0 lineto " +
				xSize + " 0 lineto " +
				xSize + " " + ySize + " lineto " +
				"closepath " +
				"clip newpath\n" +
				"1 -1 scale\n" +
				"0 -" + ySize + " translate\n");
			}
			catch (IOException ioe)
			{
				System.out.println("Error: unable to write to file: " + snapShotWriter.toString());
			}
		}
		
		//System.out.println("here");
		//drawSiteNodeDensity(gridDepth[0] > gridDepth[1] ? gridDepth[0] : gridDepth[1]);
		
	}
	
	public void drawLabelBox(String name, int fontheight, double[] cellMin, double[] cellMax, boolean centerX) {
		int rX = 0, X = 0;
		int rY = 1 ,Y = 1;
		if (!horiz) {
			X = 1;
			Y = 0;
		}
		int cellMinPix[] = {w2s(cellMin[X], X), w2s(cellMin[Y], Y)};
		int cellMaxPix[] = {w2s(cellMax[X], X), w2s(cellMax[Y], Y)};
		// draw label only when namewidth < diff[X] AND fontheight < diff[Y] in vertical direction
		int diff[] = { cellMaxPix[X] - cellMinPix[X], cellMaxPix[Y] - cellMinPix[Y]};
		int namewidth = AccordionDrawer.stringWidth(name, fontheight);
		while (namewidth > diff[X] - 3 && name.length() > 0)
		{
			name = name.substring(0, name.length()/2);
			namewidth = AccordionDrawer.stringWidth(name, fontheight);
		}
		GLFunc gl = (GLFunc) getGraphicsStuff();
		if( name.length() > 0 && fontheight <= (diff[Y] - 3)) {
//			System.out.println("nw: " + namewidth + " diffX: " + diff[X] + " fh: " + fontheight + " diffY: " + diff[Y]);
			double pos[] = {centerX ? s2w(cellMinPix[X] +(diff[X]/2.0)-(namewidth/2.0),X) : s2w(cellMinPix[X],X),//+(diff[X]/2.0)-(namewidth/2.0),X),
					 s2w(cellMinPix[Y]+(diff[Y]/2.0f)+(fontheight/2.0f),Y)};
//			System.out.println("name: " + name + "pos: " + pos[X] + " " + pos[Y]);
			drawText(pos[X], pos[Y], name, fontheight, getLabelColor(), getLabelplane(), getLabelBackColor());

		}
	}
	
	// called by something not called
	private void drawDensity(GridCell c, float depth) {
		float thecol[] = new float[3];
		float size = (float)c.getSize(X);
		float pixel = 1.0f/getWinsize(X);
		double levelsize = size;
		int level = 0;
		GridCell cell = c;
		float refpixel = (float) 10.0/getWinsize(X);
		float alpha = ( size - pixel ) / ( refpixel - pixel );
		float grayValue = (1 - alpha ) * 0.7f + alpha * 1.0f; 
		
		float diff = 0; //(c.getObjMax() - c.getObjMin()) / (float)getSequence(0).endNumOfNodes;
		// objmax and objmin were never set for any cells!

		
		//System.out.println("color " + grayValue + "counter " + depth);
		thecol[0] =  grayValue - diff;
		thecol[1] =  grayValue - diff;
		thecol[2] =  grayValue - diff;
		double zplane = colorgrid ?  - (5 + 2.0f) : hiliteplane;
		GLFunc gl = (GLFunc) getGraphicsStuff();
		gl.glColor3f(thecol[0], thecol[1], thecol[2]);
		gl.glBegin(GLEnum.GL_POLYGON);
		double[] min = {c.getMin(0),0.001f};
		double[] max = {c.getMax(0), splitLine[Y].getMinStuckValue()};
		gl.glVertex3d(min[0], max[1], zplane);
		gl.glVertex3d(min[0], min[1], zplane);
		gl.glVertex3d(max[0], min[1], zplane);
		gl.glVertex3d(max[0], max[1], zplane);
		gl.glEnd();
		//System.out.println("bbox: xmin " + min[0] + " xmax " + max[0] + " ymin " + min[1] + " ymax " + max[1]);
		//System.out.println(" R " + thecol[0] + " G " + thecol[1] + " B " + thecol[2] + " " + c.getSize(X,false) );
		}
	
	// not called
	public void drawSiteNodeDensity(float depth)
	{
		for(int nodeCount = 0; nodeCount < ((Sequence)sequences.get(0)).getNumOfNodes(); nodeCount++)
		{
			// THIS IS EXPENSIVE!!!!!!!!!!
			GridCell cell = ((Sequence)sequences.get(0)).getNode(nodeCount).getCell(); // fake cell
			
			double cellMin[] = {cell.getMin(0), 0.001f}; //cell.getMin(1, false)};
			double cellMax[] = {cell.getMax(0), splitLine[Y].getMinStuckValue()}; //cell.getMax(1, false)};
			int cellMinPix[] = {w2s(cellMin[0], 0), w2s(cellMin[1], 1)};
			int cellMaxPix[] = {w2s(cellMax[0], 0), w2s(cellMax[1], 1)};
			//drawLabelBox(name, fontHeight[1], cellMinPix, cellMaxPix);
			drawDensity(cell, depth);//drawDensity(size, )
		}

		
	}
	
	private void drawNameLabels()
	{
		if (sequences.size() == 1) { return; } 
		int[] fontHeight = {getMinFontHeight(), getMaxFontHeight()};
		int maxNameWidth = (int)(0.3 * getWinsize(X));
//		for (int fontSize = fontHeight[1]; fontSize >= fontHeight[0]; fontSize--)
//		{
			Iterator iter = listOfSequences.iterator();
			while (iter.hasNext())
			{
				SplitCell sc = (SplitCell)iter.next();
				Sequence currSeq = getSequence(sc.min+1);
				String name = currSeq.getName();
				int nameWidth = AccordionDrawer.stringWidth(name, getMaxFontHeight());
				if (nameWidth > maxNameWidth) // need to cut off some characters from name
				{
//					System.out.println("trying to chop: " + name);
					int nameResidue = nameWidth - maxNameWidth; // residue to remove in pixels
					int charsToChop = (int)((double)nameResidue/(double)nameWidth * name.length());
//					System.out.println("residue: " + nameResidue + "  charsToChop: " + charsToChop);
					if (charsToChop <= name.length() && charsToChop >= 0)
						name = name.substring(0, name.length() - charsToChop);
					else
						name = "";
				}
				if (drawlabels && name.length() > 0) {
//					System.out.println("drawing label: " + currSeq);
					double cellMin[] = {0.001f, splitLine[Y].getAbsoluteValue(sc.min, frameNum)};
					double cellMax[] = {splitLine[X].getMinStuckValue(), splitLine[Y].getAbsoluteValue(sc.max, frameNum)};
//					int cellMinPix[] = {w2s(cellMin[0], 0), w2s(cellMin[1], 1)};
//					int cellMaxPix[] = {w2s(cellMax[0], 0), w2s(cellMax[1], 1)};
					drawLabelBox(name, fontHeight[1], cellMin, cellMax, false);
				}
			}
			
//		}
	}

	protected void drawPreContFrame() {
			countDrawnFrame = 0;
	}

	protected void drawPostScene() {
//		DebugFrame.setStateMachine(DebugFrame.GL_DEPTH _TEST, false);
		gl.glDisable(GLEnum.GL_DEPTH_TEST);
//		drawPostFrame();
		if (takeSnapshot)
		{
			try
			{
				snapShotWriter.write("grestore\n" +
				"showpage\n");
				takeSnapshot = false;
				snapShotWriter.close();
			}
			catch (IOException ioe)
			{
				System.out.println("Error: can not write to file: " + snapShotWriter);
			}

		}
		// commented so now isn't needed, if needed, pass it properly, not a state machine!
//		if (dumpstats) {
//			stats.println(
//				" scene "
//					+ countDrawnScene
//					+ " frametimeCum "
//					+ cumFrameTime
//					+ " sceneTIME "
//					+ (now - continueStart));
//			//stats.flush();
//		}
	}

	protected void drawPostFrame() {
//		long now = System.currentTimeMillis();
//		countDrawnScene += countDrawnFrame;
//		long frameTime = now - dynamicStart;
//		cumFrameTime += frameTime;
//		if (dumpstats)
//			stats.print(" df " + countDrawnFrame + " frameTIME " + frameTime);
		
	}

	protected void drawBruteForce() {

	}

    public OlduvaiObject getOlduvaiObject()
    {
        return (Sequence)sequences.get(0);
    }
	
	public Sequence getSequence(int seqNum) {
		return (Sequence)sequences.get(seqNum);
	}
	public Canvas getCanvas() {
		return this;
	}
	public SiteNode getNodeForDrawing(int seqNum, int key) {
		return getSequence(seqNum).getNode(key);
	}

	public int getLabelBuffer(int xy) {
		return labelbuffer[xy];
	}
	public void setLabelBuffer(int buffer) {
		if (buffer < 1)
			buffer = 1;
		labelbuffer[X] = buffer;
		labelbuffer[Y] = buffer;
		requestRedraw();
	}
	public void setLabelBuffer(int buffer, int xy) {
		if (buffer < 1)
			buffer = 1;
		labelbuffer[xy] = buffer;
		requestRedraw();
	}
	public void increaseLabelBuffer(int xy) {
		setLabelBuffer(labelbuffer[xy] + 1, xy);
	}
	public void increaseLabelBuffer() {
		setLabelBuffer(labelbuffer[X] + 1);
	}
	public void decreaseLabelBuffer(int xy) {
		setLabelBuffer(labelbuffer[xy] - 1, xy);
	}
	public void decreaseLabelBuffer() {
		setLabelBuffer(labelbuffer[X] - 1);
	}

	public int getMaxFontHeight() {
		return maxFontHeight;
	}
	public void setMaxFontHeight(int fontheight) {
		if (fontheight < 1)
			fontheight = 1;
		maxFontHeight = fontheight;
		requestRedraw();
	}
	public void increaseMaxFontHeight() {
		setMaxFontHeight(maxFontHeight + 1);
	}
	public void decreaseMaxFontHeight() {
		setMaxFontHeight(maxFontHeight - 1);
	}

	public int getMinFontHeight() {
		return minFontHeight;
	}
	public void setMinFontHeight(int fontheight) {
		if (fontheight < 1)
			fontheight = 1;
		minFontHeight = fontheight;
		requestRedraw();
	}
	public void increaseMinFontHeight() {
		setMinFontHeight(minFontHeight + 1);
	}
	public void decreaseMinFontHeight() {
		setMinFontHeight(minFontHeight - 1);
	}

	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#subpixelDraw(AccordionDrawer.GridCell)
	 */
	public void subpixelDraw(GridCell c) {
		// TODO Auto-generated method stub
	}

	
	public void drawDensity( int minObj, int maxObj)
	{
		final float logSeqLength = (float)(1 / Math.log(getSequence(0).getNumOfNodes()) * Math.log(2));
		float thecol;
		double zplane = colorgrid ?  - (5 + 2.0f) : hiliteplane;
		//GLFunc gl = (GLFunc) getGraphicsStuff();
		
		double c1 = splitLine[X].getAbsoluteValue(minObj, frameNum);
		double c2 = splitLine[X].getAbsoluteValue(maxObj, frameNum);
//		double size = c2 - c1;
//		float white_size = (float)Math.log(getWinsize(X));  
		
		thecol = 1 - (maxObj - minObj) * logSeqLength / (float)getMinCellDims(X);
		final double densityYMin = 0.001f;
		double densityYMax = splitLine[Y].getMinStuckValue();
		
		if (takeSnapshot)
		{
			int cellMinPix[] = {w2s(c1, X), w2s(densityYMin, Y)};
			int cellMaxPix[] = {w2s(c2, X), w2s(densityYMax, Y)};
			try
			{
				snapShotWriter.write("newpath " +
				cellMinPix[X] + " " + cellMinPix[Y] + " moveto " +
				cellMinPix[X] + " " + cellMaxPix[Y] + " lineto " +
				cellMaxPix[X] + " " + cellMaxPix[Y] + " lineto " +
				cellMaxPix[X] + " " + cellMinPix[Y] + " lineto " +
				"closepath " +
				"gsave " + thecol + " " + thecol + " " + thecol + " setrgbcolor " +
				"eofill grestore\n");
			}
			catch (IOException ioe)
			{
				System.out.println("Error: IOException while trying to write cells to file: " + snapShotWriter.toString());
			}
		}
		else
		{
			gl.glColor3f(thecol, thecol, thecol);
			gl.glBegin(GLEnum.GL_POLYGON);
//			double[] min = { c1, 0.001f };
//			double[] max = { c2, splitLine[Y].getMinStuckValue()};
			gl.glVertex3d(c1, densityYMax, zplane);
			gl.glVertex3d(c1, densityYMin, zplane);
			gl.glVertex3d(c2, densityYMin, zplane);
			gl.glVertex3d(c2, densityYMax, zplane);
			gl.glEnd();
		}
	}

	public static void setGapRangeList(RangeList gapRangeList) {
		AccordionSequenceDrawer.gapRangeList = gapRangeList;
	}

	public static RangeList getGapRangeList() {
		return gapRangeList;
	}
	
	public void addToDrawQueue(Object r)
	{
		if (r != null)
			ToDrawQ.add(r);
	}
	
    public void splitRanges(int interval)
    {
		Vector newDrawQueue = new Vector();
		Iterator iter = ToDrawQ.iterator();
		while (iter.hasNext())
		{
			Object nextObj = iter.next();
			if (nextObj instanceof RangeInSequence)
			{
				RangeInSequence ris = (RangeInSequence)nextObj;
				if (ris.max - ris.min + 1 < interval)
					newDrawQueue.add(ris);
				else
				{
					int i = ris.min;
					for (i = ris.min; i < ris.max - interval; i += interval)
						newDrawQueue.add(new RangeInSequence(i, i + interval, ris.seq));
					newDrawQueue.add(new RangeInSequence(i, ris.max, ris.seq));
				}
			}
			else // maybe seed sitenodes later?  have to draw them specially
				newDrawQueue.add(nextObj);
		}
		ToDrawQ = newDrawQueue;
//		System.out.println("seeded: " + ToDrawQ.size());
	}
    
	public void splitRangeHelp(Vector drawQueue, double pixelSize, RangeInSequence ris)
	{
	    
	    Sequence seq = ris.seq;
	    SiteNode n = seq.getNode(ris.max);
	    GridCell c = n.getCell();
	    SiteNode n2 = seq.getNode(ris.min+1);
	    GridCell c2 = n2.getCell();
	    if ((ris.seq.getNode(ris.max).getCell().getMax(X) - 
		        ris.seq.getNode(ris.min+1).getCell().getMin(X) < pixelSize) ||
		     (ris.min+1 == ris.max))
	    {	
	        drawQueue.add(ris);
	        return;
	    }
	    int half = SplitLine.getThisRoot(ris.min, ris.max);
	    RangeInSequence min = new RangeInSequence(ris.min, half, ris.seq);
	    RangeInSequence max = new RangeInSequence(half, ris.max, ris.seq);
	    splitRangeHelp(drawQueue, pixelSize, min);
	    splitRangeHelp(drawQueue, pixelSize, max);
	}
    
	public void splitRanges(double pixelSize)
	{
	    Vector newDrawQueue = new Vector();
		Iterator iter = ToDrawQ.iterator();
		while (iter.hasNext())
		{
			Object nextObj = iter.next();
			if (nextObj instanceof RangeInSequence)
			{
				RangeInSequence ris = (RangeInSequence)nextObj;
			    splitRangeHelp(newDrawQueue, pixelSize, ris);
			}
			else if (nextObj instanceof SiteNode)
				newDrawQueue.add(nextObj);
		}
		ToDrawQ = newDrawQueue;
//		System.out.println("seeded: " + ToDrawQ.size() + " frame: " + frameNum);
	}
    
    public int getBottomGrid( boolean horizontal )
	{
		return ( horizontal ? bottomSize[AccordionDrawer.X] : bottomSize[AccordionDrawer.Y]) - 1;	
	}
	
	/**
	 * @return Returns the listOfSites.
	 */
	public TreeSet getListOfSites() {
		return listOfSites;
	}
};
