/*
 * Created on Dec 14, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package AccordionSequenceDrawer;

import java.util.ArrayList;

/**
 * @author jslack
 *
 * ColumnNumberList: list of ColumnNumberSets (array of tallies for each siteNode type) for all sequences
 * 
 * A full list for each sequence with the totals of each siteNode type.  This is required to compute the 
 * hierarchy of color values for larger ranges of siteNodes.  Only a few of these numberLists are left after
 * the first drawing cycle (one reason why the first draw takes a long time is the time required to create,
 * compute, destroy with these data structures, but it saves time on subsequent draws), at most half the number
 * of blocks wide.  These are replaced by coloumnColorSets when splitCell ranges (the listOfSites arrays created
 * with each draw) are altered; a larger block is computed only when required and this is not very often after
 * the first drawing cycle.  Therefore, although this datastructure is required and initially slow, it helps create
 * smaller datastructures that are sufficient and faster than this data structure for all future computations of
 * ranges of sequences. 
 */
public class ColumnNumberList {
	
	ArrayList columnNumberList;
	
	public ColumnNumberList(ColumnNumberList existingCnl, ColumnNumberList newCnl)
	{
		columnNumberList = new ArrayList();
		if (newCnl == null)
			System.out.println("Error: adding a null cnl");
		else if (existingCnl == null) // new list
			for (int i = 0; i < newCnl.columnNumberList.size(); i++)
				addNumberSet(newCnl.get(i), i);
		else // (existingCnl != null) // merge lists
			for (int i = 0; i < existingCnl.columnNumberList.size(); i++)
			{
				addNumberSet(existingCnl.get(i), i);
				addNumberSet(newCnl.get(i), i);
			}
	}
	
	public static void resetArray(int[] countArray)
	{
		for (int i = 0; i < countArray.length; i++)
			countArray[i] = 0;
	}
	
	public static void addToCount(int[] countArray, char currChar)
	{
		if (ColumnNumberSet.typeString.indexOf(currChar) != -1)
			countArray[ColumnNumberSet.typeString.indexOf(currChar)]++;
		else System.out.println("Unable to add " + currChar +
				" since it is not a known character (" + ColumnNumberSet.typeString +
				"): If you want to get rid of this message, append " + currChar +
				" to AccordionSequenceDrawer.ColumnNumberSet.typeString");
	}
	
	public ColumnNumberList(ArrayList sequences, int min, int max)
	{
		columnNumberList = new ArrayList();
		for (int currSeq = 0; currSeq < sequences.size(); currSeq++)
		{
			Sequence seq = (Sequence)sequences.get(currSeq);
			int[] typeCount = new int[ColumnNumberSet.numTypes];
			resetArray(typeCount);
			for (int i = min; i <= max; i++)
				addToCount(typeCount, seq.getNodeName(i));
			ColumnNumberSet cns = new ColumnNumberSet(null, typeCount);
			addNumberSet(cns, currSeq);
		}
	}
	
	public void addNumberSet(ColumnNumberSet cns, int index)
	{
		if (columnNumberList.size() > index)
		{
			columnNumberList.set(index, 
					new ColumnNumberSet((ColumnNumberSet)columnNumberList.get(index), cns.cumulativeCount));
		}
		else if (columnNumberList.size() == index)
			columnNumberList.add(cns);
		else
			System.out.println("Error in addNumberSet: bad index: " + index);
	}
	
	public ColumnNumberSet get(int index)
	{
		if (columnNumberList == null)
			System.out.println("This columnNumberList is not initialized properly, the array is null");
		return (ColumnNumberSet)columnNumberList.get(index);
	}
}
