/*
 * Created on Dec 14, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package AccordionSequenceDrawer;

/**
 * @author jslack
 *
 * ColumnNumberSet: array of tallies for each siteNode type
 * 
 * necessary to determine winning sitenode types for each range of siteNode.  The winning strategy
 * favours lexicographic and is not balanced since a globally balanced winning strategy causes flip-flop
 * winners (i.e. for a column of equal amounts of 'A' and 'T', we get ATATATATA...).  We perhaps need a
 * column based winning solution?  with a global entire column count on which type is favoured and a 
 * tendancy to even the global count? 
 */
public class ColumnNumberSet {

	public final static String typeString = "-ACGTUNBDHKMRSVWY"; // BDHKMRSVWY are valid, similar to N may represent several normal characters
	public final static int numTypes = typeString.length();
	private static boolean globalColumnStrategy = true;
	
	public int[] cumulativeCount;

	
	private void initializeCount(ColumnNumberSet existing, int[] incomingCount)
	{
		if (existing == null)
			cumulativeCount = incomingCount;
		else 
		{
			cumulativeCount = new int[numTypes];
			for (int i = 0; i < numTypes; i++)
				cumulativeCount[i] += incomingCount[i] + existing.cumulativeCount[i];
		}
	}
	
	public ColumnNumberSet(ColumnNumberSet existing, int[] incomingCount)
	{
		if (incomingCount.length != numTypes)
			System.out.println("badly formed incoming count to new column number set");
		else
			initializeCount(existing, incomingCount);
	}
	
	public ColumnNumberSet(ColumnColorSet existing, char letter)
	{
		if (existing != null)
			System.out.println("the existing column color set should be null in the column number set constructor");
		else
		{
			cumulativeCount = new int[numTypes];
			ColumnNumberList.resetArray(cumulativeCount);
			if (typeString.indexOf(letter) != -1) // can create an empty CNS with a different character for global count
				cumulativeCount[typeString.indexOf(letter)]++;
		}
	}
	
	public char getColor()
	{
		int currBestLetter = 0;
		int currBestCount = 0; // don't give gaps any credit, but they still win by default if nothing else there
		for (int i = 1; i < numTypes; i++)
		{
			if ((cumulativeCount[i] > currBestCount) || // i strictly wins
				(globalColumnStrategy && // turn this off and on to show the difference
						cumulativeCount[i] == currBestCount && // tie
						ColumnColorList.globalCount.cumulativeCount[i] < // in the past, i has lost more
						ColumnColorList.globalCount.cumulativeCount[currBestLetter])
				)
			{
				currBestLetter = i;
				currBestCount = cumulativeCount[i];
			}
		}
		return typeString.charAt(currBestLetter);//typeLetter[currBestLetter];
	}
	
	public String toString()
	{
		String str = "[";
		for (int i = 0; i < numTypes; i++)
			str += cumulativeCount[i] + ", ";
		return str + "]";
	}
	
	public void addColor(char letter)
	{
		cumulativeCount[typeString.indexOf(letter)]++;
	}
	
}
