
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
	 notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
	 notice, this list of conditions and the following disclaimer in the
	 documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
	 nor the names of its contributors may be used to endorse or promote
	 products derived from this Software without specific prior written
	 permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
*/

package AccordionSequenceDrawer;

import java.awt.Color;
import java.util.TreeSet;
import java.util.Iterator;

import AccordionDrawer.*;


/**
 * A class representing a list of RangeInTrees. A RangeInTree
 * represents ordered pair (min,max)) of TreeNodes.
 *
 * This class is a helper class for TreeJuxtaposer and keeps a
 * resizeable array of RangeInTree's.
 *
 * Note that each RangeInTrees item in a RangeList can be associated
 * with a different Tree.
 * 
 * @see TreeJuxtaposer.RangeInList
 * @see AccordionDrawer.TreeNode
 * 
 * @author  Tamara Munzner
 * */

public class RangeList extends AbstractRangeList {

	/**
	 * Resizeable array (linked list) of RangeInTree objects
	 */
//	  LinkedList ranges; // resizeable array of RangeInTree objects
	TreeSet ranges; // resizeable array of RangeInTree objects
	boolean enabled;
	public static boolean returnObject = false;
	public static RangeInSequence matchRange;
	Color col;
	
//	 2 numbers for the top and bottom drawing position relative to the cell height
	double[] drawingRange = new double[2];
	
	int key;
    
	/*
	 * Only highlight group items for this tree, not all the others
	 */ 
	boolean thisTreeOnly;

	public RangeList(int i) {
	col = new Color((float).5, (float).5, (float).5);
	enabled = true;
	ranges = new TreeSet();
	thisTreeOnly = false;
	key = i;
	}
	public RangeList(int i, Color c, boolean on, boolean only) {
	key = i;
	col = c;
	enabled = on;
	ranges = new TreeSet();
	thisTreeOnly = only;
	}
    
	public void addRange(int min, int max, Sequence seq, boolean draw, boolean collideDetect)
	{ // add range with intersection checking, this means the list is ordered
		RangeInSequence addSequence = new RangeInSequence(min, max, seq); // new range to add
		RangeInSequence intersectRange;
		if (collideDetect)
		{
			RangeInSequence.checkSequence = !draw; // true => sequence matters (marking), false => any sequence is ok
			RangeList.returnObject = true;
			do
			{
				RangeList.matchRange = null;
				if (ranges.contains(addSequence))
				{
					if (addSequence.min > RangeList.matchRange.min)
						addSequence.min = RangeList.matchRange.min;
					if (addSequence.max < RangeList.matchRange.max)
						addSequence.max = RangeList.matchRange.max;
					ranges.remove(RangeList.matchRange);
				}
			}
			while (RangeList.matchRange != null);
			// reset
			RangeList.returnObject = false;
			RangeInSequence.checkSequence = true;
	}
	insertInOrder(addSequence, draw);
}

	// this doesn't work
	public void removeRange(int min, int max, Sequence seq) {
		System.err.println("can't remove a range like this");
	RangeInSequence r = new RangeInSequence(min, max, seq);
	// delete if it it's identical to something previously passed in...
	ranges.remove(r);
	}

	// treeset version
	private void insertInOrder(RangeInSequence seq, boolean draw)
	{
		ranges.add(seq);
	}

	public void clear() {
	ranges.clear();
	}

	void print() {
	RangeInSequence rt;
	try{
	RangeList temp = (RangeList)this.clone();
	for (int i = 0; i < ranges.size(); i++) {
		rt = (RangeInSequence)temp.ranges.first();
		temp.ranges.remove(rt);
		System.out.println(i + " treekey " + rt.seq.getKey() + " min " + rt.min + " max " + rt.max);
	}
	}
	catch (CloneNotSupportedException e)
	{
		System.out.println("Cloning should work for rangeLists");
		return;
	}

	}

	public boolean getEnabled() {return enabled;}
	public void setDrawingRange(double[] drawingRange) {this.drawingRange[0] = drawingRange[0]; this.drawingRange[1] = drawingRange[1];}
	public double[] getDrawingRange() {return drawingRange;}
	public void setEnabled(boolean on) { enabled = on;}
	public void setColor(Color c) {col = new Color(c.getRed(), c.getGreen(), c.getBlue(), key);}
	public Color getColor() {if (col != null && col.getAlpha() != key) col = new Color(col.getRed(), col.getGreen(), col.getBlue(), key); return col;}
	public float getDepth(float hiliteplane) {return hiliteplane - .004f*key;}
	public boolean getThisTreeOnly() { return thisTreeOnly;}
	void setThisTreeOnly(boolean on) { thisTreeOnly = on;}
	public int getKey() { return key; }
	public int size() { return ranges.size(); }
	public double[] getSizesOfAllRanges(SplitLine splitLine, int frameNum)
	{
		Iterator iter = ranges.iterator();
		double[] size = new double[size()];
		int i = 0;
		while (iter.hasNext())
		{
			RangeInSequence range = (RangeInSequence)iter.next();
			size[i++] = range.getSize(splitLine, frameNum);
		}
		return size;
	}
//	double noShrink = group.getUnshrinkableTotal(this, frameNum); // total of all shrinking regions that will not shrink
	public double getUnshrinkableTotal(AccordionDrawer ad, SplitLine splitLine, int frameNum)
	{
		double size = 0.0;
		if (ranges.size() == 0) return size;
		Iterator iter = ranges.iterator();
		RangeInSequence ris = (RangeInSequence)iter.next();
		RangeInSequence betweenRanges = new RangeInSequence(0, ris.min, ris.seq);
		double betweenSize = betweenRanges.getSize(splitLine, frameNum);
		if (betweenSize < ad.minContextPeriphery)
			size += betweenSize;
		int prevMax = ris.max;
		while (iter.hasNext())
		{
			ris = (RangeInSequence)iter.next();
			if (iter.hasNext())
			{
				betweenRanges = new RangeInSequence(prevMax, ris.min, ris.seq);
				betweenSize = betweenRanges.getSize(splitLine, frameNum);
				if (betweenSize < ad.minContextInside)
					size += betweenSize;
			}
			prevMax = ris.max;
		}
		betweenRanges = new RangeInSequence(ris.max, splitLine.getSize(), ris.seq);
		betweenSize = betweenRanges.getSize(splitLine, frameNum);
		if (betweenSize < ad.minContextPeriphery)
			size += betweenSize;
		return size;
	}
	public String toString()
	{
		String returnString = key+"[";
		Iterator iter = ranges.iterator();
		RangeInSequence next = null;
		if (iter.hasNext())
			next =(RangeInSequence)iter.next(); 
		while (next != null && iter.hasNext())
		{
			returnString += next + ",\n\t";
			next = (RangeInSequence)iter.next();
		}
		if (ranges.size() > 0)
			returnString += next;
		returnString += "]";
		return returnString;
	}

	/**
	 * @return
	 */
	public TreeSet getRanges() {
		return ranges;
	}
	
	public RangeInSequence getHead() {
		return (RangeInSequence)ranges.first();
	}

// treeset can't support this function efficiently	
//	public RangeInSequence getRange(int i){
//		return (RangeInSequence)ranges.get(i);
//	}


	/**
	 * @param i
	 */
	public void setKey(int i) {
		key = i;
	}

	/**
	 * @param list
	 */
	public void setRanges(TreeSet list) {
		ranges = list;
	}
	
	public int getNumRanges()
	{
		return ranges.size();
	}

	
	/* (non-Javadoc)
	 * @see AccordionDrawer.RangeList#getSplitIndices(boolean)
	 */
	// return splitindices for each of the ranges, min and max, in order
	public int[] getSplitIndices(boolean horizontal) {
		int xy = horizontal ? AccordionSequenceDrawer.X : AccordionSequenceDrawer.Y;
		int[] returnArray = new int[ranges.size() * 2];
		Iterator iter = ranges.iterator();
		int i = 0;
		while (iter.hasNext())
		{
			RangeInSequence currRange = (RangeInSequence)iter.next(); 
			returnArray[i*2] = currRange.min-1;
			returnArray[i*2+1] = currRange.max;
			i++;
		}
		return returnArray;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AbstractRangeList#onlyThisAD(AccordionDrawer.AccordionDrawer)
	 */
	public AbstractRangeList onlyThisAD(AccordionDrawer d) {
		return this; // everything is in the same AD
	}

	// Ranges that call this must be in a single sequence only, i.e sequences with
	//  ranges that overlap in columns are not good to use here, only things that we
	//  use to grow can be used here, search results might have to be tweaked to
	//  have a draw group as well as the mark gorup
	public AbstractRangeList flipRangeToShrink(int xy, AccordionDrawer ad)
	{
		RangeList returnRangeList = new RangeList(this.key);
		int splitLineSize = ad.getSplitLine(xy).getSize()-1; 
		if (ranges.size() == 0)
		{
			return null;
		}
		Iterator flipIter = ranges.iterator();
		RangeInSequence prev = (RangeInSequence)flipIter.next();
		if (prev.min > 0)
			returnRangeList.addRange(0, prev.min-1, prev.seq, false, false);
		while( flipIter.hasNext())
		{
			RangeInSequence curr = (RangeInSequence)flipIter.next();
		if (prev.seq != curr.seq)
			System.out.println("Error, sequences don't match: " + curr.seq + " " + prev.seq);
		// do something
		returnRangeList.addRange( prev.max+1, curr.min-1, prev.seq, false, false );
		prev = curr; 
	}
	if (prev.max < splitLineSize)
		returnRangeList.addRange(prev.max+1, splitLineSize, prev.seq, false, false);
	return returnRangeList;
}
	
    public RangeInSequence returnRangeOverlap(RangeInSequence ris, boolean checkSeqNum)
    {
		RangeInSequence returnRange = null;
		RangeInSequence.checkSequence = checkSeqNum;
		matchRange = null;
		returnObject = true;
		if (ranges.contains(ris) && matchRange != null)
			returnRange = matchRange;
		matchRange = null;
		returnObject = false;
		RangeInSequence.checkSequence = true;
		return returnRange;
	}
	
	public void cutRange(RangeInSequence ris)
	{
		RangeInSequence matchingRange = returnRangeOverlap(ris, false);
		if (matchingRange != null) {
			ranges.remove(matchingRange);
			if (ris.min > matchingRange.min)
				addRange(matchingRange.min, ris.min - 1, ris.seq, false, false);
			if (ris.max < matchingRange.max)
				addRange(ris.max + 1, matchingRange.max, ris.seq, false, false);
		}
	}
	
//	RangeList blockForest = new RangeList(group, asd.listOfSites, siteSplits);
	public RangeList(RangeList group, TreeSet siteList, SplitLine siteSplits)
	{
		Iterator iter = group.ranges.iterator();
		ranges = new TreeSet();
		while (iter.hasNext())
		{
			RangeInSequence risOrig = (RangeInSequence)iter.next();
			SplitCell scBlock = SplitCell.getOverlapSplitCell(siteList, new SplitCell(risOrig.min-1, risOrig.max));
			int scBlockMinCell = scBlock.min+1;
			if (risOrig.max - risOrig.min < scBlock.max - scBlockMinCell)
			{
				int risOrigCenter = (risOrig.min + risOrig.max) / 2; // don't use center of block, use center of marked group
				int dist[] = {risOrigCenter - scBlockMinCell, scBlock.max - risOrigCenter};
				int growSize = dist[0] > dist[1] ? dist[0] : dist[1]; // max distance should be beyond risOrig.min,max from center
				ranges.add(new RangeInSequence(risOrigCenter - growSize, risOrigCenter + growSize, risOrig.seq));
			}
			else
				ranges.add(risOrig); // block for this range is too small
		}
	}


};

// at some point, consider whether we should deal with overlaps intelligently.
// what does this statement mean?  I know I wrote it, but it confuses me now (2004-06-22, james)
