/*
 * Created on 20-Feb-2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package SeqParser;

import AccordionSequenceDrawer.*;
import SequenceJuxtaposer.*;
import java.io.*;
import java.util.*;

/**
 * @author hilde
 *
 * Parser for Fasta Files 
 * Right now, we don't check the alphabet. That means
 * every character is accepted
 * only files get accepted that look like this:
 * 
 * 	> name of the sequence
 * 	MYFUNNYDNA?RNASTRIng
 * 	> next sequence
 * 	UNTIL THE END
 * 
 * SequenceStrings are disassembled in strings a SequenceJuxtaposer.MAX_STRING_LEN size
 * 
 */
public class SequenceLoader {
	
	int key = 0;
	boolean eof = false; // class wide EOF
	
	public SequenceLoader( SequenceJuxtaposer sj, String filename, int maxLoad) throws FileNotFoundException
	{
		try {
				BufferedReader br = new BufferedReader(new FileReader(filename));
				String line = "";
				ArrayList seqStrings = new ArrayList();
				String seqPart = "";
				Sequence seq; 
				while( !eof && key < maxLoad )
				{
					seq = readSequence( br );
//					System.out.println(" seq " + seq + " " + seq.getKey());
					sj.addSequence(seq);
				}
				sj.createASD(); // here do all gap processing 	
							
		} catch (IOException e) {
			System.err.println("this is not a DNA sequence " + e.getMessage());
		}

		
	}
	
	/*
	 * reads from the filedescriptor
	 */
	public Sequence readSequence( BufferedReader br ) throws IOException
	{
			String line = br.readLine();
			String seqName;
			ArrayList seqStrings; 
			
			// check if file is correct and and !EOF				
			if (line == null) 
			{
				  throw new IOException("EOF");
			}
			while(line.length() == 0) 
			{
			  	line = br.readLine();
			  	if (line == null) 
			  	{
						throw new IOException("EOF");
			  	}
			}
			
			// get Sequence description
			if (!line.startsWith(">")) 
			{
			  	throw new IOException("File doesn't contain FASTA formatted data: " + line);
			}
    					
			line = line.substring(1); // cut off '>' and trim
			seqName = line.trim();
			System.out.println("added " + seqName);
			
			seqStrings = readSequenceData( br );
			
			if( seqStrings == null )
				return null;
			else
				return new Sequence( seqName, seqStrings, key++ );
	}
	
	/*
	 * read Sequence data and put it in an ArrayList
	 * adopted from biojava.org, fastaformat.java
	 */
	public ArrayList readSequenceData( BufferedReader br ) throws IOException
	{
		char[] cache = new char[512];
		ArrayList seqStrings = new ArrayList();
		String seqPart = "";
		StringBuffer sb = new StringBuffer();
		int len = 0;
		boolean end = false;
		
		// read til next sequence
		while( !end )
		{
			br.mark( cache.length + 1 );
			int bytesRead = br.read( cache, 0, cache.length );
			if ( bytesRead < 0 ) 
			{
					end = true;
					eof = true;
			} 
			else 
			{
				int parseStart = 0;
				int parseEnd = 0;
				while ( !end && parseStart < bytesRead && cache[parseStart] != '>' ) 
				{
					  parseEnd = parseStart;
          
					  while (parseEnd < bytesRead && cache[parseEnd] != '\n' && cache[parseEnd] != '\r' ) 
					  {
						++parseEnd;
					  }

					  sb.append( cache, parseStart, parseEnd - parseStart );
					  len = sb.length();
					  if (  len >= SequenceJuxtaposer.MAX_STRING_LEN ) 
					  {
							seqStrings.add( sb.substring( 0, SequenceJuxtaposer.MAX_STRING_LEN ).toUpperCase() );
					  		sb.delete( 0, SequenceJuxtaposer.MAX_STRING_LEN );
					  }
					 
					  parseStart = parseEnd + 1;
					  while (parseStart < bytesRead && (cache[parseStart] == '\n' || cache[parseStart] == '\r') )
						{
						  ++parseStart;
						}
								
				}
				
				
				if (parseStart < bytesRead && cache[parseStart] == '>') 
				{
					try 
					{ 
				  		br.reset();
					} 
					catch (IOException e) 
					{
						throw new IOException( "Can't reset: " +  
						e.getMessage() + " parseStart=" + 
						parseStart + " bytesRead=" + 
						bytesRead);
					}
					if ( br.skip(parseStart) != parseStart ) 
					{
							throw new IOException("Couldn't reset to start of next sequence");
					}
					end = true;
				}
				
		 	}
			//put the rest as an arraylist at the end 
			if( sb.length() > 0 && end ) {
				seqStrings.add( sb.substring( 0, sb.length() ).toUpperCase() );
				sb.delete( 0, sb.length() );
			}	
		}
		
		// split String in single strings and add them
//		System.out.println("string arrays in seq " + seqStrings.size());				
		return seqStrings;	
	}
	
	
	

}
